from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List

import typer
from fyeah import f
from pynmrstar import Saveframe

from nef_pipelines.lib.nef_frames_lib import CCPN_MERIT, SPECTRUM_FRAME_CATEGORY
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    extract_column,
    read_or_create_entry_exit_error_on_bad_file,
    set_column,
    set_column_to_value,
)
from nef_pipelines.lib.sequence_lib import MoleculeTypes, sequence_from_entry
from nef_pipelines.lib.util import (  # STDIN,; exit_error,; parse_comma_separated_options,
    STDIN,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.echidna import import_app
from nef_pipelines.transcoders.sparky.importers.peaks import (
    pipe as sparky_peak_import_pipe,
)

DEFAULT_MERIT_FUNTION = "round(x**(1.0/6.0),5)"

# TODO: rationalise these into lib
CCPN_COMMENT = "ccpn_comment"
HEIGHT = "height"
NEF_PEAK = "nef_peak"

app = typer.Typer()

DEFAULT_NUCLEI_HELP = (
    "nuclei to use for each dimension, if not defined they are guessed from the assignments"
    "or an error is reported"
)


# noinspection PyUnusedLocal
class BadEchidnaFileException(Exception):
    pass


@dataclass
class TensorFrame:
    amplitude: float
    rhombicity: float

    theta: float
    psi: float
    phi: float


@dataclass
class Position:
    x: float
    y: float
    z: float


TENSOR_HEADERS = set("Xax Xrh theta psi phi".split())
ATOM_POSITION_HEADERS = set("-x -y -z".split())
TENSOR_AND_ATOM_HEADERS = [*TENSOR_HEADERS, *ATOM_POSITION_HEADERS]


def _parse_echidna(input: Iterator[str], file_name: str):

    lines = []
    number_column_counts = OrderedDict()

    in_data = False
    tensor_header_indices = {}
    for line_no, line in enumerate(input):
        line = line.strip()
        fields = line.split()

        if line_no == 0:
            _check_for_echidna_header_and_raise_if_bad(fields, file_name, line)
            continue

        if line_no == 1:

            fields = [field.replace("metal", "") for field in fields]

            _check_for_echidna_tensor_and_raise_if_bad(fields, file_name, line)

            _check_for_correct_number_tensor_fields_and_raise_if_bad(
                fields, file_name, line
            )

            tensor_headers = fields[2:]

            _check_tensor_headings_and_raise_if_bad(tensor_headers, line, file_name)

            tensor_header_indices = {
                tensor_header: index
                for index, tensor_header in enumerate(tensor_headers)
            }

            continue

        if line_no == 2:

            _check_tensor_values_in_comment_or_raise(fields, line, file_name)

            tensor_and_atom_position = fields[1:]

            NUM_TENSOR_FIELDS = len(TENSOR_AND_ATOM_HEADERS)
            _check_number_tensor_fields_or_raise(
                len(tensor_and_atom_position), NUM_TENSOR_FIELDS, line, file_name
            )

            tensor = _parse_tensor_values_and_raise_if_bad(
                fields, tensor_header_indices, file_name, line
            )
            atom_position = _parse_atom_position_and_raise_if_bad(
                fields, tensor_header_indices, file_name, line
            )
            continue

        if len(fields) == 0:
            continue

        if line_no > 2 and fields[0] == "Assignment":
            in_data = True
            continue

        if line_no > 2 and in_data:
            number_count = 0
            for field in fields[1:]:
                if is_float(field):
                    number_count += 1

            number_column_counts[(line, line_no)] = number_count
            lines.append(line)

    column_count = set(number_column_counts.values())
    if len(column_count) > 1:

        raise BadEchidnaFileException()

    else:
        num_dimension_columns = list(number_column_counts.values())[0] - 1
        dimensions = " ".join([f"w{i+1}" for i in range(num_dimension_columns)])
        header = f"Assignment {dimensions} Data Height"
        lines.insert(0, "")
        lines.insert(0, header)

    return lines, tensor, atom_position


def _parse_tensor_values_and_raise_if_bad(
    fields, tensor_header_indices, file_name, line
):
    FIELD_TRANSLAIONS = {"Xax": "amplitude", "Xrh": "rhombicity"}

    fields = fields[1:]  # to remove the #

    values = {}
    for field_name in TENSOR_HEADERS:
        value = fields[tensor_header_indices[field_name]]

        if not is_float(value):
            msg = f"""
                The tensor field {field_name} in the file {file_name} is not a floating point number
                the value was {value}

                the line was

                {line}
            """
            raise BadEchidnaFileException(msg)
        else:
            field_name = (
                FIELD_TRANSLAIONS[field_name]
                if field_name in FIELD_TRANSLAIONS
                else field_name
            )
            values[field_name] = float(value)

    return TensorFrame(**values)


def _parse_atom_position_and_raise_if_bad(
    fields, tensor_header_indices, file_name, line
):

    FIELD_TRANSLAIONS = {"-x": "x", "-y": "y", "-z": "z"}

    values = {}
    for field_name in ATOM_POSITION_HEADERS:
        value = fields[tensor_header_indices[field_name]]
        if not is_float(value):
            msg = f"""
                The the atom position field {field_name} in the file {file_name} is not a floating point number
                the value was {value}

                the line was

                {line}
            """
            raise BadEchidnaFileException(msg)
        else:

            values[FIELD_TRANSLAIONS[field_name]] = float(value)

    return Position(**values)


def _check_number_tensor_fields_or_raise(
    number_fields, NUM_TENSOR_FIELDS, line, file_name
):
    if number_fields != NUM_TENSOR_FIELDS:
        msg = f"""
            in the file {file_name} at line 2 for an echidna file I expected there to be
            {NUM_TENSOR_FIELDS} fields I got {number_fields}

            the line was

            {line}
        """
        raise BadEchidnaFileException(msg)


def _check_tensor_values_in_comment_or_raise(fields, line, file_name):
    if len(fields) == 0 or fields[0] != "#":
        msg = f"""
                    An echidna tensor headers values must be in a comment, file {file_name} line 2 did not start
                    with a comment I got

                    {line}
                """
        raise BadEchidnaFileException(msg)


def _check_tensor_headings_and_raise_if_bad(tensor_headers, line, file_name):

    if set(tensor_headers) != set(TENSOR_AND_ATOM_HEADERS):
        msg = f"""
                    in the file {file_name} at line 2 for an echidna file I expected the tensor headers
                    to include: {' '.join(TENSOR_AND_ATOM_HEADERS)}
                    i got:  {' '.join(tensor_headers)}

                    the line was:  {line}

                    please note the text 'metal' in the line is ignored
                """
        raise BadEchidnaFileException(msg)


def _check_for_correct_number_tensor_fields_and_raise_if_bad(fields, file_name, line):

    if len(fields) != 10:
        msg = f"""
                In an echidna file I expected to have 10 tensor values in the tensor header [line 2 of the file]:

                    # Tensor: Xax, Xrh, theta, psi, phi, -x, -y, -z

                in the file {file_name} I got {len(fields) - 2}  after the comment character [#]

                the line was:

                {line}

                please note the text 'metal' has been filtered out from the input line if it was present
            """
        raise BadEchidnaFileException(msg)


def _check_for_echidna_tensor_and_raise_if_bad(fields, file_name, line):
    if fields[0] != "#" and fields[1] != "Tensor":
        msg = f"""
                        in the file {file_name} at line 1 for an echidna file I expected '# Tensor' i got
                        {line}
                    """
        raise BadEchidnaFileException(msg)


def _check_for_echidna_header_and_raise_if_bad(fields, file_name, line):
    if fields[0] != "#" and fields[1] != "Echidna":
        msg = f"""
                    in the file {file_name} at line 0 for an echidna file I expected '# Echidna' i got
                    {line}
                """
        raise BadEchidnaFileException(msg)


HELP_MERIT_FUNCTION = """
    function to calculate a merit value from an echidna error value, the error value is provided as the parameter x
"""


@import_app.command(no_args_is_help=True)
def peaks(
    frame_name: str = typer.Option(
        "{file_name}",
        "-f",
        "--frame-name",
        help="a templated name for the frame {file_name} will be replaced by input filename without its extension",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    nuclei: List[str] = typer.Option([], help=DEFAULT_NUCLEI_HELP),
    molecule_type: MoleculeTypes = typer.Option(
        MoleculeTypes.PROTEIN, help="the type of molecule"
    ),
    default_chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain-code",
        help="default chain code to use if none is provided in the file",
    ),
    no_validate: bool = typer.Option(
        False,
        help="if set don't validate the peaks agains the inpuy sequence if provided",
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
    merit_function: str = typer.Option(DEFAULT_MERIT_FUNTION, help=HELP_MERIT_FUNCTION),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type peaks.txt", metavar="<ECHIDNA-peaks>.txt"
    ),
):
    """convert echidna peak assignment file <ECHIDNA-PEAKS>.txt to NEF"""

    file_data = {}
    for file_name in file_names:
        with open(file_name) as fh:

            file_data[file_name] = fh.read().split("\n")

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    sequence = sequence_from_entry(entry) if not no_validate else None

    nuclei = parse_comma_separated_options(nuclei)

    for file_name, lines in file_data.items():
        entry = pipe(
            entry,
            frame_name,
            {file_name: lines},
            default_chain_code,
            sequence,
            input_dimensions=nuclei,
            spectrometer_frequency=spectrometer_frequency,
            molecule_type=molecule_type,
            merit_function=merit_function,
        )

    print(entry)


def is_iterable(target):
    result = True
    try:
        iter(target)
    except TypeError:
        result = False

    return result


#
def pipe(
    entry,
    frame_name_template,
    file_names_and_lines,
    chain_code,
    sequence,
    input_dimensions,
    spectrometer_frequency,
    molecule_type=MoleculeTypes.PROTEIN,
    merit_function=DEFAULT_MERIT_FUNTION,
):

    for file_name, lines in file_names_and_lines.items():

        lines, tensor, atom_position = _parse_echidna(lines, file_name)

        entry = sparky_peak_import_pipe(
            entry,
            frame_name_template,
            {file_name: lines},
            chain_code,
            sequence,
            input_dimensions,
            spectrometer_frequency,
            molecule_type=molecule_type,
        )

        file_name = file_name.stem  # used by f()

        frame_code = f"{SPECTRUM_FRAME_CATEGORY}_{f(frame_name_template)}"

        frame = entry.get_saveframe_by_name(frame_code)

        frame_name = file_name

        loop = frame.get_loop(NEF_PEAK)

        loop.add_tag(CCPN_MERIT, update_data=True)

        comment_values = _remove_and_store_comments(loop)

        height_index = loop.tag_index(HEIGHT)
        merit_index = loop.tag_index(CCPN_MERIT)

        _convert_heights_to_figure_of_merit(
            loop, height_index, merit_index, merit_function
        )

        comment_index = _append_stored_comments(loop, comment_values)

        _add_frame_comment_for_large_violation_limit(
            frame, loop, comment_index, merit_index
        )

        _add_tensor_frame_saveframe(entry, frame, tensor, frame_name)

        _add_atom_position_saveframe(entry, frame, atom_position, frame_name)

    return entry


def _add_atom_position_saveframe(entry, frame, atom_position, frame_name):
    atom_position_category = "np_atom_position"
    atom_position_name = f"{atom_position_category}_{frame_name}"
    frame.add_tag("np_atom_position_name", atom_position_name)
    tag_values = {"x": atom_position.x, "y": atom_position.y, "z": atom_position.z}
    atom_position_save_frame = create_simple_frame(
        atom_position_category, atom_position_name, tag_values
    )
    entry.add_saveframe(atom_position_save_frame)


def _add_tensor_frame_saveframe(entry, frame, tensor, frame_name):
    tensor_frame_category = "np_tensor_frame"
    tensor_frame_name = f"{tensor_frame_category}_{frame_name}"
    frame.add_tag("np_tensor_frame_name", tensor_frame_name)
    tag_values = {
        "restraint_origin": "measured",
        "ccpn_format": "angles_euler",
        "restraint_magnitude": tensor.amplitude,
        "restraint_rhomicity": tensor.rhombicity,
        "ccpn_phi": tensor.phi,
        "ccpn_psi": tensor.psi,
        "ccpn_theta": tensor.theta,
    }
    tensor_save_frame = create_simple_frame(
        tensor_frame_category, tensor_frame_name, tag_values
    )
    entry.add_saveframe(tensor_save_frame)


def _add_frame_comment_for_large_violation_limit(
    frame, loop, comment_index, merit_index
):
    merits_for_large_violations = []
    for row in loop:
        if "large violation!" in row[comment_index]:
            merits_for_large_violations.append(float(row[merit_index]))
    lowest_large_violation_merit = min(merits_for_large_violations)
    note = f"NOTE: merits with values > {lowest_large_violation_merit} are flagged as large violations"
    frame.add_tag(CCPN_COMMENT, note)


def _append_stored_comments(loop, comment_values):
    loop.add_tag(CCPN_COMMENT, update_data=True)
    comment_index = loop.tag_index(CCPN_COMMENT)
    set_column(loop, comment_index, comment_values)
    return comment_index


def _convert_heights_to_figure_of_merit(
    loop, height_index, merit_index, merit_function
):
    heights = extract_column(loop, height_index)
    heights = [float(height) for height in heights]
    set_column_to_value(loop, height_index, UNUSED)
    merits = [eval(merit_function) for x in heights]
    set_column(loop, merit_index, merits)


def _remove_and_store_comments(loop):
    comment_index = loop.tag_index(CCPN_COMMENT)
    comment_values = extract_column(loop, comment_index)
    comment_values = [
        comment.lstrip().lstrip(";").lstrip() for comment in comment_values
    ]
    loop.remove_tag(CCPN_COMMENT)
    return comment_values


def create_simple_frame(atom_position_category, atom_position_name, tag_values):
    save_frame = Saveframe.from_scratch(atom_position_name, atom_position_category)

    save_frame.add_tag("sf_category", atom_position_category)
    save_frame.add_tag("sf_framecode", atom_position_name)
    for tag, value in tag_values.items():
        save_frame.add_tag(tag, value)

    return save_frame



import typer
from fyeah import f

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.talos.importers.restraints import restraints

app = typer.Typer()
app.command()(restraints)

EXPECTED = """
save_nef_dihedral_restraint_list_talos_restraints_A_PHI_PSI                                                # noqa: E501
   _nef_dihedral_restraint_list.sf_category     nef_dihedral_restraint_list
   _nef_dihedral_restraint_list.sf_framecode    nef_dihedral_restraint_list_talos_restraints_A_PHI_PSI
   _nef_dihedral_restraint_list.potential_type  square-well-parabolic
   _nef_dihedral_restraint_list.ccpn_comment    'file: {pred_4_path}, source file: ubiq.tab, talos version: 4.12 Rev 2023.083.11.36'

   loop_
      _nef_dihedral_restraint.index
      _nef_dihedral_restraint.restraint_id
      _nef_dihedral_restraint.restraint_combination_id
      _nef_dihedral_restraint.name
      _nef_dihedral_restraint.chain_code_1
      _nef_dihedral_restraint.sequence_code_1
      _nef_dihedral_restraint.residue_name_1
      _nef_dihedral_restraint.atom_name_1
      _nef_dihedral_restraint.chain_code_2
      _nef_dihedral_restraint.sequence_code_2
      _nef_dihedral_restraint.residue_name_2
      _nef_dihedral_restraint.atom_name_2
      _nef_dihedral_restraint.chain_code_3
      _nef_dihedral_restraint.sequence_code_3
      _nef_dihedral_restraint.residue_name_3
      _nef_dihedral_restraint.atom_name_3
      _nef_dihedral_restraint.chain_code_4
      _nef_dihedral_restraint.sequence_code_4
      _nef_dihedral_restraint.residue_name_4
      _nef_dihedral_restraint.atom_name_4
      _nef_dihedral_restraint.target_value
      _nef_dihedral_restraint.target_value_error
      _nef_dihedral_restraint.ccpn_comment
      _nef_dihedral_restraint.np_merit

     1   1   .   PHI   A   3   ILE   C   A   2   GLN   N    A   2   GLN   CA   A   2   GLN   C   -101.524   15.07    'class: Dyn'        0.0
     2   2   .   PSI   A   2   GLN   N   A   2   GLN   CA   A   2   GLN   C    A   3   ILE   N    139.087    15.435   'class: Dyn'        0.0
     3   3   .   PHI   A   4   PHE   C   A   3   ILE   N    A   3   ILE   CA   A   3   ILE   C   -145.993   11.455   'class: Strong'     1.0
     4   4   .   PSI   A   3   ILE   N   A   3   ILE   CA   A   3   ILE   C    A   4   PHE   N    159.239    9.155    'class: Strong'     1.0
     5   5   .   PHI   A   5   VAL   C   A   4   PHE   N    A   4   PHE   CA   A   4   PHE   C   -118.216   7.883    'class: Generous'   0.6
     6   6   .   PSI   A   4   PHE   N   A   4   PHE   CA   A   4   PHE   C    A   5   VAL   N    137.017    9.702    'class: Generous'   0.6

   stop_

save_
""".replace(
    NOQA_E501, ""
)


def test_pred4tab():

    pred_4_path = path_in_test_data(__file__, "pred_4.tab")
    pred_4_nef_path = path_in_test_data(__file__, "pred_4_seq.nef")

    STREAM = read_test_data(pred_4_nef_path)

    result = run_and_report(
        app,
        [
            pred_4_path,
        ],
        input=STREAM,
    )

    phi_psi = isolate_frame(
        result.stdout, "nef_dihedral_restraint_list_talos_restraints_A_PHI_PSI"
    )

    PATCHED_EXPECTED = f(EXPECTED)
    assert_lines_match(PATCHED_EXPECTED, phi_psi)


EXPECTED_CHI1 = """
save_nef_dihedral_restraint_list_talos_restraints_A_CHI1                                                   # noqa: E501
   _nef_dihedral_restraint_list.sf_category     nef_dihedral_restraint_list
   _nef_dihedral_restraint_list.sf_framecode    nef_dihedral_restraint_list_talos_restraints_A_CHI1
   _nef_dihedral_restraint_list.potential_type  square-well-parabolic
   _nef_dihedral_restraint_list.ccpn_comment    'file: {pred_4_path}, source file: ubiq.tab, talos version: 4.12 Rev 2023.083.11.36'

   loop_
      _nef_dihedral_restraint.index
      _nef_dihedral_restraint.restraint_id
      _nef_dihedral_restraint.restraint_combination_id
      _nef_dihedral_restraint.name
      _nef_dihedral_restraint.chain_code_1
      _nef_dihedral_restraint.sequence_code_1
      _nef_dihedral_restraint.residue_name_1
      _nef_dihedral_restraint.atom_name_1
      _nef_dihedral_restraint.chain_code_2
      _nef_dihedral_restraint.sequence_code_2
      _nef_dihedral_restraint.residue_name_2
      _nef_dihedral_restraint.atom_name_2
      _nef_dihedral_restraint.chain_code_3
      _nef_dihedral_restraint.sequence_code_3
      _nef_dihedral_restraint.residue_name_3
      _nef_dihedral_restraint.atom_name_3
      _nef_dihedral_restraint.chain_code_4
      _nef_dihedral_restraint.sequence_code_4
      _nef_dihedral_restraint.residue_name_4
      _nef_dihedral_restraint.atom_name_4
      _nef_dihedral_restraint.target_value
      _nef_dihedral_restraint.target_value_error
      _nef_dihedral_restraint.ccpn_comment
      _nef_dihedral_restraint.np_merit

     1   1   .   CHI1   A   3   ILE   N   A   3   ILE   CA   A   3   ILE   CB   A   3   ILE   CG1   61.9    6.8    'Q_Gm: 0.087, Q_Gt: 0.768, Q_T: 0.145'   1.0
     2   2   .   CHI1   A   4   PHE   N   A   4   PHE   CA   A   4   PHE   CB   A   4   PHE   CG   -61.1   10.8   'Q_Gm: 0.644, Q_Gt: 0.307, Q_T: 0.049'   1.0

   stop_

save_

""".replace(
    NOQA_E501, ""
)


def test_pred4tab_chi1():

    pred_4_path = path_in_test_data(__file__, "pred_4_chi1.tab")
    pred_4_nef_path = path_in_test_data(__file__, "pred_4_seq.nef")

    STREAM = read_test_data(pred_4_nef_path)

    result = run_and_report(
        app,
        [
            pred_4_path,
        ],
        input=STREAM,
    )

    phi_psi = isolate_frame(
        result.stdout, "nef_dihedral_restraint_list_talos_restraints_A_CHI1"
    )

    PATCHED_EXPECTED = f(EXPECTED_CHI1)
    assert_lines_match(PATCHED_EXPECTED, phi_psi)


# TODO check filtering & bad inputs



from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.transcoders.nmrpipe.importers.sequence import pipe as nmrpipe_pipe
from nef_pipelines.transcoders.talos import import_app

app = typer.Typer()


@import_app.command()
def sequence(
    chain_code: str = typer.Option(
        "A",
        "--chain",
        help="chain codes for the new chain",
        metavar="<CHAIN-CODE>",
    ),
    no_chain_start: bool = typer.Option(
        False,
        "--no-chain-start/",
        help="don't include a start of chain link type for the first residue",
    ),
    no_chain_end: bool = typer.Option(
        False,
        "--no-chain-end/",
        help="don't include an end of chain link type for the last residue",
    ),
    entry_name: str = typer.Option("talos", help="a name for the entry"),
    input: Path = typer.Option(
        None,
        "-i",
        "--input",
        metavar="|INPUT|",
        help="file to read NEF data from [- is stdin; defaults is stdin]",
    ),
    file_name: Path = typer.Argument(
        ..., help="input files of type <TALOS-FILE>.pred", metavar="<PRED-FILE>"
    ),
):
    """-  convert sequence from a talos pred file to NEF"""

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name=entry_name)

    lines = read_file_or_exit(file_name)

    entry = pipe(entry, lines, chain_code, no_chain_start, no_chain_end, file_name)

    print(entry)


def pipe(
    entry: Entry,
    lines: List[str],
    chain_code: str,
    no_chain_start: bool,
    no_chain_end: bool,
    file_name: Path,
):

    return nmrpipe_pipe(
        entry, lines, chain_code, no_chain_start, no_chain_end, 1, file_name
    )



import math
from dataclasses import dataclass, field
from enum import auto
from typing import Dict, List, Optional, Tuple, Union

from strenum import LowercaseStrEnum, StrEnum

# TODO: to avoid circular import, move to constants
UNUSED = "."


class Linking(StrEnum):
    START = auto()
    MIDDLE = auto()
    END = auto()
    FREE = auto()


class SequenceResidue: ...  # noqa: E701


class Residue: ...  # noqa: E701


# something like this might be nice
# but we would need updates...
# class ResidueAssignmentState(StrEnum):
#     UNASSIGNED = auto()
#     ASSIGNED = auto()
#     OFFSET_OR_PREFIX = auto()
#     LABEL = auto()


@dataclass(frozen=True, order=True)
class Residue:
    chain_code: str
    sequence_code: Union[int, str]
    residue_name: str
    sequence_code_prefix: str = ""
    offset: int = 0
    # assignmnet_state:ResidueAssignmentState  add as property

    @staticmethod
    def from_sequence_residue(sequence_residue: SequenceResidue) -> Residue:
        return Residue(
            sequence_residue.chain_code,
            sequence_residue.sequence_code,
            sequence_residue.residue_name,
        )


@dataclass(frozen=True, order=True)
class SequenceResidue(Residue):

    is_cis: bool = False
    linking: Optional[Linking] = None
    variants: List[str] = ()


# should contain a residue and have constructors?
@dataclass(frozen=True, order=True)
class AtomLabel:
    residue: Residue
    atom_name: str
    element: str = None
    isotope_number: int = None

    def is_unassigned(self):
        residue_unassigned = (
            self.residue.chain_code == UNUSED
            and self.residue.sequence_code == UNUSED
            and self.residue.residue_name == UNUSED
        )
        return residue_unassigned and self.atom_name == UNUSED


UNASSIGNED_ATOM = AtomLabel(Residue(UNUSED, UNUSED, UNUSED), UNUSED)


@dataclass
class PeakAxis:
    atom_labels: List[AtomLabel]
    ppm: float
    merit: str
    # comment: str


@dataclass
class DistanceRestraint:
    atom_list_1: List[AtomLabel]
    atom_list_2: List[AtomLabel]

    target_distance: float
    distance_minus: float
    distance_plus: float

    comment: str = None


@dataclass
class DihedralRestraint:
    atom_1: AtomLabel
    atom_2: AtomLabel
    atom_3: AtomLabel
    atom_4: AtomLabel

    merit: float = None
    name: str = None
    remark: str = None

    target_value: float = None  # use one or the  of target_value and error
    target_value_error: float = None

    lower_limit: float = None  # or upper and lower limits
    upper_limit: float = None


@dataclass
class PeakValues:
    serial: int

    height: Optional[float] = None
    height_uncertainty: Optional[float] = None

    volume: Optional[float] = None
    volume_uncertainty: Optional[float] = None

    deleted: Optional[bool] = False
    comment: Optional[str] = ""

    width: Optional[float] = None  # HWHH ppm
    # bound: float

    # merit: Optional[str] = None

    # flag0: str


# assignment has a tuple of dimensions


@dataclass
class Assignments:
    assignments: Dict[str, List[AtomLabel]]


@dataclass
class Peak:
    id: int
    values: PeakValues

    # move these to axis_values?
    positions: Dict[str, float]

    # assignment has a list of one or more assignments
    # each Assignment will have one value for each axis this maybe be either
    # 0. a list with no AtomLabels - unassigned
    # 1. a list with a single AtomLabel -  this axis is definitively assigned
    # 2. a list with multiple AtomLabels - this axis has multiple putative assignments
    # Note if there are multiple unique assignments each of these is should be a top level
    # assignment of the peak
    assignments: List[Assignments]

    position_uncertainties: Optional[Dict[str, float]] = None


@dataclass
class PeakListData:
    num_axis: int
    axis_labels: List[str]
    # isotopes: List[str]
    data_set: str
    sweep_widths: List[float]
    spectrometer_frequencies: List[float]


# TODO: are axes indexed by names or by integer...
@dataclass
class PeakList:
    peak_list_data: PeakListData
    peaks: List[Dict[Union[int, str], Peak]]


@dataclass
class LineInfo:
    file_name: str
    line_no: int
    line: str


@dataclass(frozen=True, order=True)
class ShiftData:
    atom: AtomLabel
    value: float  # TODO: should be position
    value_uncertainty: Optional[float] = None  # TODO: should be position_uncertainty
    line_width: Optional[float] = None  # line width in Hz
    line_width_uncertainty: Optional[float] = None  # uncertainty of line width in Hz

    frame_name: Optional[str] = ""
    frame_row: Optional[int] = None
    frame_line: Optional[str] = None


@dataclass
class ShiftList:
    shifts: List[ShiftData]


@dataclass(order=True)
class RdcRestraint:
    atom_1: AtomLabel
    atom_2: AtomLabel
    value: float
    value_uncertainty: float
    weight: Optional[float] = None


class PeakFitMethod(LowercaseStrEnum):
    GAUSSIAN = auto()
    LORENTZIAN = auto()
    SPLINE = auto()


@dataclass(frozen=True, order=True)
class NewPeak:
    # TODO: support multiple assignments
    shifts: List[
        ShiftData
    ]  # multiple assignments  we support this by having mutiple shifts with the same value?

    id: Optional[int] = None
    height: Optional[float] = None
    height_uncertainty: Optional[float] = None
    volume: Optional[float] = None
    volume_uncertainty: Optional[float] = None
    peak_fit_method: Optional[Union[str, PeakFitMethod]] = None
    figure_of_merit: Optional[float] = None
    comment: str = ""


@dataclass
class DimensionInfo:
    axis_code: str  # this is the isotope code for us...
    axis_name: str = None
    axis_unit: Optional[str] = "ppm"


class RelaxationModelParameter(StrEnum):
    S2 = auto()
    S2_FAST = auto()
    S2_SLOW = auto()
    TAU_E = auto()
    TAU_FAST = auto()
    TAU_SLOW = auto()
    R_EXCHANGE = auto()
    J_0 = auto()
    J_OMEGA_1 = auto()
    J_OMEGA_2 = auto()
    J_OMEGA_1_87 = auto()
    J_OMEGA_2_87 = auto()
    RHO_1 = auto()
    RHO_2 = auto()


class RelaxationModelType(StrEnum):
    MODEL_FREE = auto()
    REDUCED_SPECTRAL_DENSITY = auto()


class RelaxationUnit(StrEnum):
    UNITLESS = auto()
    PICO_SECOND = auto()
    NANO_SECOND = auto()
    MICRO_SECOND = auto()
    MILLI_SECOND = auto()
    PER_SECOND = auto()
    SECOND = auto()


class RelaxationDataSource(StrEnum):
    SIMULATION = auto()
    ESTIMATE = auto()
    EXPERIMENTAL = auto()


@dataclass(frozen=True, order=True)
class RelaxationValue:
    atom: AtomLabel
    value: float
    value_type: RelaxationModelParameter
    value_error: float = None
    unit: RelaxationUnit = None
    dipole_atom: AtomLabel = None


class DiffusionModel(StrEnum):
    SPHERE = auto()
    SPHEROID_PROLATE = auto()
    SPHEROID_OBLATE = auto()
    ELLIPSOID = auto()


# when we get beyond 3.9 support this should change to
# Vector3D: TypeAlias = Tuple[float, float, float]
# and when we get beyon 3.11 support this should change to
# type Vector3D = Tuple[float, float, float]
Vector3D = Tuple[float, float, float]


@dataclass(frozen=True)
class RdcTensorFrameData:
    da: float
    dr: float

    eigen_vector_x: Vector3D
    eigen_vector_y: Vector3D
    eigen_vector_z: Vector3D

    def is_defined(self):
        da_ok = not math.isnan(self.da)
        dr_ok = not math.isnan(self.dr)
        eigen_x_ok = not any(math.isnan(elem) for elem in self.eigen_vector_x)
        eigen_y_ok = not any(math.isnan(elem) for elem in self.eigen_vector_y)
        eigen_z_ok = not any(math.isnan(elem) for elem in self.eigen_vector_z)
        eigen_ok = eigen_x_ok and eigen_y_ok and eigen_z_ok

        return da_ok and dr_ok and eigen_ok


@dataclass(frozen=True)
class TensorFrame:
    d_iso: float = None

    d_x: float = None
    d_y: float = None
    d_z: float = None

    d_anisotropic: float = None
    d_rhombic: float = None

    alpha: float = None
    beta: float = None
    gamma: float = None

    theta: float = None
    phi: float = None


@dataclass(frozen=True)
class RelaxationData:
    model_type: RelaxationModelType
    data_source: RelaxationDataSource
    values: List[RelaxationValue] = field(default_factory=list)
    tauM: float = None
    tauM_error: float = None

    field_strengths: List[float] = None

    diffusion_model: DiffusionModel = None
    tensor_frame: TensorFrame = None
    tensor_frame_error: TensorFrame = None
    structure_name: str = None


# fmt: off
# TODO: this is an initial attempt some more work needed on turn types and other analysis programs
class SecondaryStructureType(LowercaseStrEnum):
    #                                         DSSP |  SST |  comment
    ALPHA_HELIX = auto()                    # H    |  H   |  right-handed
    ALPHA_HELIX_LEFT_HANDED = auto()        # .    |  h
    THREE_TEN_HELIX = auto()                # G    |  G   |  right-handed
    THREE_TEN_HELIX_LEFT_HANDED = auto()    # .    |  g
    BETA_SHEET = auto()                     # E    |  E
    BETA_BRIDGE = auto()                    # B    |
    PI_HELIX = auto()                       # I    |  I   |  right-handed, aka a 5 helix
    PI_HELIX_LEFT_HANDED = auto()           # .    |  i
    TURN = auto()                           # T    |  T   |  hydrogen bonded turn, subclasses listed below
    ALPHA_TURN = auto()                     # .    |  4   |  4 bonds, ALPHA-like TURN (right/left-handed)
    BETA_TURN = auto()                      # .    |  3   |  3 bonds sub classes are listed below
    #                                              |      |  3_10-like  TURN (right/left-handed)
    BETA_I_TURN = auto()
    BETA_II_TURN = auto()
    BETA_I_PRIME_TURN = auto()
    BETA_II_PRIME_TURN = auto()
    BETA_IV_TURN = auto()
    BETA_VIA1_TURN = auto()
    BETA_VIA2_TURN = auto()
    BETA_VIB_TURN = auto()
    BETA_VII_TURN = auto()
    GAMMA_TURN = auto()                     # .    |      |  2 bonds
    DELTA_TURN = auto()                     # .    |      |  1 bond - sterically unlikley
    PI_TURN = auto()                        # .    |   5  |  5 bonds
    BEND = auto()                           # S    |      |
    COIL = auto()                           # .    |      |  either disordered or a loop but unknown which
    LOOP = auto()                           # .    |      |  well-defined structure but not one of the
    #                                              |      |  standard dssp types also called an omega loop
    DISORDERED = auto()                     # .    |      |  a region without ordered secondary structure
    UNKNOWN = auto()                        # .    |   -  |  the program wasn't able to determine secondary
    #                                              |      |  structure type (None?)
# fmt: on


@dataclass
class SecondaryStructure:
    residue: Residue
    secondary_structure: SecondaryStructureType
    merit: float  # typically between 0 and 1
    comment: str = ""



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:
    nef_app.app.add_typer(
        app,
        name="xplor",
        help="- read xplor [sequences, dihedral & distance restraints]",
    )

    app.add_typer(
        import_app,
        name="import",
        help="- import xplor [sequences, dihedral & distance restraints]",
    )

    app.add_typer(
        export_app,
        name="export",
        help="- export xplor [rdcs]",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.xplor.exporters.rdcs  # noqa: F401
    import nef_pipelines.transcoders.xplor.importers.dihedrals
    import nef_pipelines.transcoders.xplor.importers.distances  # noqa: F401
    import nef_pipelines.transcoders.xplor.importers.sequence  # noqa: F401



from pathlib import Path
from typing import List

from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.sequence_lib import (
    get_chain_starts,
    get_residue_name_from_lookup,
    offset_chain_residues,
    sequence_3let_list_from_sequence,
    sequence_3let_to_sequence_residues,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    DiffusionModel,
    RelaxationData,
    RelaxationDataSource,
    RelaxationModelParameter,
    RelaxationModelType,
    RelaxationUnit,
    RelaxationValue,
    Residue,
    SecondaryStructure,
    SecondaryStructureType,
)
from nef_pipelines.lib.util import convert_to_float_or_exit, exit_error
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    VALUES,
    DbFile,
    dbfile_to_first_residue_number,
    exit_if_required_columns_missing,
    gdb_to_3let_sequence,
    get_column_indices,
    get_gdb_columns,
    select_records,
)

SECONDARY_PREDICTION_TRANSLATIONS = {"h": "H", "e": "E", "c": "L"}


SECONDARY_STUCTURE_TRANSLATION = {
    "H": SecondaryStructureType.ALPHA_HELIX,
    "E": SecondaryStructureType.BETA_SHEET,
    "L": SecondaryStructureType.COIL,
    "h": SecondaryStructureType.UNKNOWN,
    "e": SecondaryStructureType.UNKNOWN,
    "c": SecondaryStructureType.UNKNOWN,
}


def _exit_if_sequences_dont_match(chain_code, nef_sequence, talos_sequence):

    nef_sequence_set = set(
        [Residue.from_sequence_residue(residue) for residue in nef_sequence]
    )
    talos_sequence_set = set(
        [Residue.from_sequence_residue(residue) for residue in talos_sequence]
    )

    if not talos_sequence_set.issubset(nef_sequence_set):
        nef_sequence_start = get_chain_starts(nef_sequence)[chain_code]
        nef_sequence_for_chain = [
            residue for residue in nef_sequence if residue.chain_code == chain_code
        ]
        nef_sequence_list = "\n".join(
            sequence_3let_list_from_sequence(
                nef_sequence_for_chain, chain_code=chain_code
            )
        )

        talos_sequence_start = get_chain_starts(talos_sequence)[chain_code]
        talos_sequence_for_chain = [
            residue for residue in talos_sequence if residue.chain_code == chain_code
        ]
        talos_sequence_list = "\n".join(
            sequence_3let_list_from_sequence(
                talos_sequence_for_chain, chain_code=chain_code
            )
        )

        msg = f"""
                nef and talos sequences don't agree...

                talos sequence for chain {chain_code} starts at {talos_sequence_start} and has the sequence
                {talos_sequence_list}

                nef sequence for chain {chain_code} starts at {nef_sequence_start} and has the sequence
                {nef_sequence_list}
            """

        exit_error(msg)


def read_talos_secondary_structure(
    gdb_records: DbFile,
    chain_code: str,
    file_name: str,
    nef_sequence: List[Residue],
    include_predictions: bool,
) -> List[SecondaryStructure]:
    columns = get_gdb_columns(gdb_records)

    required_columns = set("RESID RESNAME CONFIDENCE SS_CLASS".split())
    if not required_columns.issubset(columns):
        exit_if_required_columns_missing(gdb_records, required_columns, file_name)

    sequence_3let = gdb_to_3let_sequence(gdb_records)

    if len(sequence_3let) > 0:

        chain_start = dbfile_to_first_residue_number(gdb_records)

        sequence = sequence_3let_to_sequence_residues(sequence_3let, chain_code)

        sequence = offset_chain_residues(sequence, {chain_code: chain_start - 1})

        _exit_if_sequences_dont_match(chain_code, nef_sequence, sequence)
    else:
        sequence = nef_sequence

    exit_error_if_no_sequence(sequence, file_name)

    return gdb_records_to_secondary_structure(
        gdb_records, sequence, chain_code, include_predictions
    )


def gdb_records_to_secondary_structure(
    gdb_records: DbFile,
    sequence: List[Residue],
    chain_code: str,
    include_predictions: bool = False,
) -> List[SecondaryStructure]:

    residue_lookup = sequence_to_residue_name_lookup(sequence)

    column_indices = get_column_indices(gdb_records)

    result = []
    for record in select_records(gdb_records, VALUES):

        sequence_code = record.values[column_indices["RESID"]]

        residue_name_3let_talos = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_lookup
        )

        talos_secondary_structure = record.values[column_indices["SS_CLASS"]]

        is_prediction = False
        if (
            include_predictions
            and talos_secondary_structure in SECONDARY_PREDICTION_TRANSLATIONS
        ):
            talos_secondary_structure = SECONDARY_PREDICTION_TRANSLATIONS[
                talos_secondary_structure
            ]
            is_prediction = True

        if talos_secondary_structure in SECONDARY_STUCTURE_TRANSLATION:
            secondary_structure = SECONDARY_STUCTURE_TRANSLATION[
                talos_secondary_structure
            ]
        else:
            line_info = record.line_info
            msg = f"""
                    the secondary structure type {talos_secondary_structure} is unrecognised at
                    line: {line_info.line_no} in
                    file: {line_info.file_name} the line was
                    line: {line_info.line}
                """
            exit_error(msg)

        if secondary_structure != SecondaryStructureType.UNKNOWN:
            merit = record.values[column_indices["CONFIDENCE"]]
            merit = convert_to_float_or_exit(merit, record.line_info)
        else:
            merit = UNUSED

        residue = Residue(
            chain_code=chain_code,
            sequence_code=sequence_code,
            residue_name=residue_name_3let_talos,
        )

        comment = ""

        if is_prediction:
            comment = "*PREDICTION*"
        value = SecondaryStructure(residue, secondary_structure, merit, comment=comment)
        result.append(value)

    return result


def gdb_records_to_s2(gdb_records, sequence, chain_code):

    residue_lookup = sequence_to_residue_name_lookup(sequence)

    column_indices = get_column_indices(gdb_records)

    values = []
    for record in select_records(gdb_records, VALUES):

        sequence_code = record.values[column_indices["RESID"]]

        residue_name_3let_talos = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_lookup
        )
        s2 = record.values[column_indices["S2"]]

        residue = Residue(
            chain_code=chain_code,
            sequence_code=sequence_code,
            residue_name=residue_name_3let_talos,
        )
        atom = AtomLabel(atom_name="N", residue=residue, element="N", isotope_number=15)
        dipole_atom = AtomLabel(
            atom_name="H", residue=residue, element="H", isotope_number=1
        )
        value = RelaxationValue(
            atom=atom,
            dipole_atom=dipole_atom,
            unit=RelaxationUnit.UNITLESS,
            value_type=RelaxationModelParameter.S2,
            value=s2,
        )
        values.append(value)

    return RelaxationData(
        model_type=RelaxationModelType.MODEL_FREE,
        values=values,
        diffusion_model=DiffusionModel.SPHERE,
        data_source=RelaxationDataSource.ESTIMATE,
    )


def read_order_parmeters(gdb_records, chain_code, file_name, nef_sequence):

    required_columns = set("RESID RESNAME S2".split())
    exit_if_required_columns_missing(gdb_records, required_columns, file_name)

    sequence_3let = gdb_to_3let_sequence(gdb_records)

    if len(sequence_3let) > 0:

        chain_start = dbfile_to_first_residue_number(gdb_records)

        sequence = sequence_3let_to_sequence_residues(sequence_3let, chain_code)

        sequence = offset_chain_residues(sequence, {chain_code: chain_start - 1})

        _exit_if_sequences_dont_match(chain_code, nef_sequence, sequence)
    else:
        sequence = nef_sequence

    exit_error_if_no_sequence(sequence, file_name)

    return gdb_records_to_s2(gdb_records, sequence, chain_code)


def exit_error_if_no_sequence(sequence, file_name):
    if len(sequence) == 0:
        pref_file_path = Path(file_name).parent / "pred.tab"
        msg = f"""
            There is no sequence from the file {file_name}, please provide a sequence as a nef input stream

            e.g.   nef talos import sequence {pref_file_path}
                 | nef talos import order-parameters {file_name}
        """
        exit_error(msg)



import re
import sys
from itertools import zip_longest
from pathlib import Path
from typing import List, Tuple

import typer
from fyeah import f
from pynmrstar import Entry, cnmrstar

from nef_pipelines import nef_app
from nef_pipelines.lib.util import (
    STDIN,
    STDOUT,
    exit_error,
    parse_comma_separated_options,
    read_from_file_or_exit,
)

FILE_PATHS_HELP = """write the entries to files named, these maybe comma separated
                    or multiple calls to the option maybe made. If a single directory is provided,
                    the entries will be saved in the directory with the entry name from the template
                    if stdout [-] is provided, the entries will be printed to stdout with delimiter
                    the default is write entries to the current directory again with the entry name
                    from the template"""

if nef_app:
    # noinspection PyUnusedLocal
    @nef_app.app.command()
    def save(
        input: Path = typer.Option(
            STDIN,
            "-i",
            "--in",
            metavar="NEF-FILE-STREAM",
            help="read NEF data stream from a file instead of stdin",
        ),
        template: str = typer.Option(
            "{entry_id}.nef",
            help="the new name for the entry",
        ),
        force: bool = typer.Option(
            False, "-f", "--force", help="overwrite existing files"
        ),
        single_file: bool = typer.Option(
            False, help="write all entries to a single file"
        ),
        no_globals_cleanup: bool = typer.Option(
            False, "--no-globals-cleanup", help="do not remove the globals frame"
        ),
        no_header: bool = typer.Option(False, help="do not write a header"),
        file_paths: List[str] = typer.Argument(None, help=FILE_PATHS_HELP),
    ):
        """- save the entries in the stream to a file / files or stdout with delimiters"""

        # this is a gratuitous hack Typer doesn't seem to allow an argument to be - so we have to check
        # TODO: report this!
        if not file_paths:
            if sys.argv[-1] == "-":
                file_paths = ["-"]
            else:
                file_paths = [
                    Path.cwd(),
                ]

        file_paths = parse_comma_separated_options(file_paths)

        if file_paths == [
            str(STDIN),
        ]:
            file_paths = [
                STDIN,
            ]
        else:
            file_paths = [Path(file_path).resolve() for file_path in file_paths]

        entries = list(_entry_list_from_stdin_or_exit_error_if_none(input))

        if not no_globals_cleanup:
            for entry in entries:
                for save_frame in entry.get_saveframes_by_category("nefpls_globals"):
                    entry.remove_saveframe(save_frame)

        entries = pipe(entries, file_paths, template, no_header, single_file, force)

        if entries:
            for entry in entries:
                print(entry)


def pipe(
    entries: List[Entry],
    file_paths: List[Path],
    template: str,
    no_header: bool,
    single_file: bool,
    force: bool,
):

    _exit_if_no_file_paths(file_paths)

    _exit_if_no_entries(entries)

    _exit_if_single_file_and_out_is_directory(file_paths, single_file)

    _exit_if_single_file_and_out_is_multiple_files(file_paths, single_file)

    # are we writing to STDOUT?
    write_stdout = file_paths == [
        STDOUT,
    ]

    DELIMITER = "-" * 50  # noqa: F841
    HEADER = "{DELIMITER} {entry_id} {DELIMITER}"

    # do we need to write a header?
    write_header = False
    write_header = not no_header and (
        file_paths
        == [
            STDOUT,
        ]
        or single_file
    )

    # every file is written to stdout if we write to STDOUT
    if write_stdout:
        single_file = True

    # make file paths from directory and entry_id id we are writing to a directory
    if len(file_paths) == 1 and file_paths[0].is_dir():
        directory = file_paths[0]
        file_paths = []
        for entry in entries:
            entry_id = entry.entry_id
            file_name = f(template)
            file_paths.append(directory / file_name)

    # deal with what append mode to use
    if single_file:
        append_modes = [
            True,
        ] * len(entries)
        append_modes[0] = False
    else:
        append_modes = [
            False,
        ] * len(entries)

    if single_file:
        file_paths = file_paths * len(entries)

    for i, (entry, file_path, append_mode) in enumerate(
        zip_longest(entries, file_paths, append_modes, fillvalue=None)
    ):

        _exit_if_we_ran_out_of_entries(entry, file_path)

        _exit_if_we_ran_out_of_file_paths(entry, file_path)

        _exit_if_file_exists_and_no_append_or_force(file_path, append_mode, force)

        file_h = (
            sys.stdout
            if file_path == STDOUT
            else open(file_path, "a" if append_mode else "w")
        )

        if len(entries) > 1 and write_header:
            entry_id = entry.entry_id  # noqa: F841
            header_text = f(HEADER)
            print(header_text, file=file_h)
        else:
            header_text = None

        print(entry, file=file_h)

        if not file_h == sys.stdout:
            file_h.close()

    if write_header and header_text:
        file_h = (
            sys.stdout
            if file_path == STDOUT
            else open(file_path, "a" if append_mode else "w")
        )
        if len(entries) > 1:
            print("-" * len(header_text), file=file_h)

        if not write_stdout:
            file_h.close()

    return None if write_stdout else entries


def _exit_if_file_exists_and_no_append_or_force(file_path, append_mode, force):

    if file_path != STDOUT and file_path.exists() and not (force or append_mode):
        msg = f"""
            when trying to write the file {file_path} it already exists,
            if you want to overwrite it use the force option
        """

        exit_error(msg)


def _exit_if_we_ran_out_of_file_paths(entry, file_path):
    if not file_path:
        msg = f"""
            Number of entries does not match number of file paths, at entry
            {entry.entry_id} i ran out of file paths to write to
        """

        exit_error(msg)


def _exit_if_we_ran_out_of_entries(entry, file_path):
    if not entry:
        msg = f"""
            Number of entries does not match number of file paths, at file name
            {file_path} i ran out of entries to write
        """

        exit_error(msg)


# stolen from pynmrstar parser
def _load_data(data: str) -> None:
    """Loads data in preparation of parsing and cleans up newlines
    and massages the data to make parsing work properly when multi-line
    values aren't as expected. Useful for manually getting tokens from
    the parser."""

    # Fix DOS line endings
    data = data.replace("\r\n", "\n").replace("\r", "\n")
    # Change '\n; data ' started multi-lines to '\n;\ndata'
    data = re.sub(r"\n;([^\n]+?)\n", r"\n;\n\1\n", data)

    cnmrstar.load_string(data)


def _get_token() -> str:
    """Returns the next token in the parsing process."""

    try:
        token, line_number, delimiter = cnmrstar.get_token_full()
    except ValueError as err:
        raise Exception(str(err))

    return token, line_number, delimiter


def _iterate_tokens() -> Tuple[str, int, str]:
    while token := _get_token():
        if token[0] is None:
            break
        yield token


def _entry_list_from_stdin_or_exit_error_if_none(file_name):

    file_path = Path(file_name)

    text = read_from_file_or_exit(file_path)

    _load_data(text)

    entry_strings = []
    tokens = None
    for token in _iterate_tokens():

        if token[0].startswith("data_"):
            if tokens:
                entry_strings.append(" ".join(tokens))
            tokens = []

        tokens.append(f"\n{token[2]}{token[0]}\n{token[2]}")

    if tokens:
        entry_strings.append(" ".join(tokens))

    return [Entry.from_string(entry_string) for entry_string in entry_strings]


def _exit_if_single_file_and_out_is_multiple_files(file_paths, single_file):
    if single_file and len(file_paths) > 1:
        file_paths_string = "\n".join([str(file_path) for file_path in file_paths])
        msg = f"""
            the single file option is incompatible with multiple file paths, there were {len(file_paths)} file paths
            the paths were
            {file_paths_string}
        """
        exit_error(msg)


def _exit_if_single_file_and_out_is_directory(file_paths, single_file):
    if single_file and file_paths[0].is_dir():
        msg = f"""
            ouput to a single file option is incompatible with specifiying a directory as the file path
            the file path you provided was {file_paths[0]}
        """
        exit_error(msg)


def _exit_if_no_file_paths(file_paths):
    if not file_paths:
        msg = "you must provide at least one file_path to write to"
        exit_error(msg)


def _exit_if_no_entries(entries):
    if not entries:
        msg = "you must provide at least one entry to write, did you input a NEF file stream?"
        exit_error(msg)



import os
import re
import sys
import tempfile
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from textwrap import dedent
from typing import Optional
from urllib.parse import urlsplit

import requests
import typer
from bs4 import BeautifulSoup
from bs4.element import Comment
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    SelectionType,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import sequences_from_frames, translate_1_to_3
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import AtomLabel, Residue, ShiftData, ShiftList
from nef_pipelines.lib.util import STDIN, exit_error, is_int
from nef_pipelines.tools.loops.trim import ChainBound
from nef_pipelines.tools.loops.trim import pipe as trim
from nef_pipelines.transcoders.shiftx2 import import_app

NETWORK_200_OK = 200

PDB_UNIPROT_MAPPING_URL_TEMPLATE = (
    "https://www.ebi.ac.uk/pdbe/api/mappings/uniprot/{code_or_filename}"
)
ALPHA_FOLD_KEY = "AIzaSyCeurAJz7ZGjPQUtEaerUkBZ3TaBkXrY94"
ALPHA_FOLD_URL_TEMPLATE = (
    "https://alphafold.ebi.ac.uk/api/prediction/{uniprot_id}?key={alphafold_key}"
)

SHIFTX2_RETRY_COUNT = 10


@dataclass
class NetworkResult:
    network_ok: bool


@dataclass
class PdbAndUniprotIDs:
    uniprot_id: str
    pdb_code: str
    chain_code: str


@dataclass
class PbdUniprotChainMapping:
    pdb_uniprot_start: Optional[int]
    pdb_uniprot_end: Optional[int]
    pdb_start_residue: Optional[int]

    @property
    def pdb_uniprot_length(self):
        return self.pdb_uniprot_end - self.pdb_uniprot_start


@dataclass
class PdbUniprotMapping(NetworkResult, PdbAndUniprotIDs, PbdUniprotChainMapping):
    raw_data: Optional[dict]


@dataclass
class AlphafoldResult(NetworkResult, PdbAndUniprotIDs, PbdUniprotChainMapping):
    pdb_url: Optional[str]
    alphafold_id: Optional[str]
    raw_data: Optional[dict]


@dataclass
class PDBDownloadResult(NetworkResult, PdbAndUniprotIDs, PbdUniprotChainMapping):
    file_system_ok: bool
    pdb_url: str
    alphafold_id: str
    pdb_file_path: Optional[Path]


def tag_visible(element):
    if element.parent.name in [
        "style",
        "script",
        "head",
        "title",
        "meta",
        "[document]",
    ]:
        return False
    if isinstance(element, Comment):
        return False
    return True


deuterate = {False: 1, True: 0}
phospho = {False: 1, True: 0}
shifttype = {"all": "back-bone", 1: 1, "side-chain": 2}
output = {"tabular": 0, 1: "csv", 2: "nmr-star", 3: "nef"}
use_shifty = {False: 1, True: 0}
multiple_chains = {False: 1, True: 0}

ROOT_URL = "http://www.shiftx2.ca/cgi-bin"
CGI_URL = f"{ROOT_URL}/shiftx2.cgi"

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    code_or_file_name: str = typer.Argument(
        None,
        help="file name to read shift data from or alphafold / pdb code to fetch data for",
    ),
    source_chain: str = typer.Option(
        None, help="chain in the source coordinate file to predict shift data for"
    ),
    alphafold: bool = typer.Option(
        False, "--alphafold", help="use alphafold to predict structure"
    ),
    chain: str = typer.Option(
        None, "-c", "--chain", help="chain to label output shift data with"
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="verbose output"),
    in_file: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="input to read NEF data from [- is stdin]",
    ),
):
    """- read a shiftx2 chemical shift prediction [alpha]"""
    entry = read_or_create_entry_exit_error_on_bad_file(in_file, "shiftx2")
    entry = pipe(entry, code_or_file_name, source_chain, chain, alphafold, verbose)
    print(entry)


def pipe(
    entry: Entry,
    code_or_filename: str,
    source_chain: str,
    chain: str,
    alphafold: bool,
    verbose: bool,
) -> Entry:

    file_path = Path(code_or_filename)
    if file_path.exists():
        shifts = _read_shifts_from_file(file_path, chain)
    else:

        if alphafold:
            pdb_file_info = _pdb_code_to_alphafold_pdb_file(
                code_or_filename, source_chain, verbose
            )
            code_or_filename = pdb_file_info.pdb_file_path
            use_file = True
        else:
            use_file = False

        if not chain and source_chain:
            chain = source_chain

        for i in range(1, SHIFTX2_RETRY_COUNT + 1):
            shifts = _get_shifts_from_server(
                code_or_filename, source_chain, chain, use_file=use_file
            )
            if shifts:
                break
            elif verbose:
                print(f"retrying shiftx2 ...[{i}]", file=sys.stderr)

        _exit_if_too_many_attempted_connections(
            shifts, code_or_filename, SHIFTX2_RETRY_COUNT
        )

    shift_list = ShiftList(shifts)
    frame = shifts_to_nef_frame(shift_list, "shiftx2")

    entry.add_saveframe(frame)

    if not chain and source_chain:
        chain = source_chain

    if not chain and not source_chain:
        chain = "A"

    if alphafold:
        sequence = sequences_from_frames(frame, chain)
        shiftx2_uniprot_start = min(
            [
                residue.sequence_code
                for residue in sequence
                if is_int(residue.sequence_code)
            ]
        )
        shiftx2_uniprot_end = max(
            [
                residue.sequence_code
                for residue in sequence
                if is_int(residue.sequence_code)
            ]
        )

        run_trim = False
        if pdb_file_info.pdb_uniprot_start > shiftx2_uniprot_start:
            start = pdb_file_info.pdb_uniprot_start
            run_trim = True
        else:
            start = shiftx2_uniprot_start

        if pdb_file_info.pdb_uniprot_start < shiftx2_uniprot_end:
            end = pdb_file_info.pdb_uniprot_end
            run_trim = True
        else:
            end = shiftx2_uniprot_end

        if run_trim:
            chain_bounds = [ChainBound(chain, start, end)]

            chain_bounds = {chain: chain_bounds}

            entry = trim(entry, "shiftx2", SelectionType.NAME, chain_bounds)

    return entry


def _exit_if_too_many_attempted_connections(shifts, code_or_filename, RETRY_COUNT):
    if not shifts:
        msg = f"""
            couldn't get a shiftx2 prediction for {code_or_filename} after {RETRY_COUNT} retries
            """
        exit_error(msg)


class ShiftFormat(Enum):
    TABULAR = auto()
    CSV = auto()
    NMRSTAR = auto()
    NEF = auto()


def _is_nef_header(line):
    line_fields = line.split() if line else []
    return len(line_fields) == 6 and line_fields[-1] == "."


def _is_star_header(line):
    line_fields = line.split() if line else []
    return len(line_fields) == 7 and line_fields[-1] == "1"


def _read_shifts_from_file(file_path, cli_chain_code):

    with open(file_path, "r") as file:
        lines = file.read()
        shifts = _parse_text_to_shifts(lines, cli_chain_code, file_path)

    return shifts


def _parse_text_to_shifts(text, cli_chain_code, source):

    lines = []
    for i, line in enumerate(text.split("\n")):
        if i == 0:
            if "BACKBONE ATOMS" in line:
                file_format = ShiftFormat.TABULAR
            elif "," in line:
                file_format = ShiftFormat.CSV
            elif _is_nef_header(line):
                file_format = ShiftFormat.NEF
            elif _is_star_header(line):
                file_format = ShiftFormat.NMRSTAR
            else:
                msg = f"""
                        When reading shiftx2 shifts could not determine format of source {source}
                        the first line in the file was
                        {line}
                        """
                exit_error(msg)
        lines.append(line)

    columns = {
        ShiftFormat.CSV: {
            "chain_code": None,
            "sequence_code": 0,
            "residue_name": 1,
            "atom_name": 2,
            "shift": 3,
        },
        ShiftFormat.NMRSTAR: {
            "chain_code": None,
            "sequence_code": 1,
            "residue_name": 2,
            "atom_name": 3,
            "shift": 4,
        },
        ShiftFormat.NEF: {
            "chain_code": 0,
            "sequence_code": 1,
            "residue_name": 2,
            "atom_name": 3,
            "shift": 4,
        },
        ShiftFormat.TABULAR: {
            "chain_code": None,
            "sequence_code": 0,
            "residue_name": 1,
            "atom_name": 2,
            "shift": 3,
        },
    }
    if file_format == ShiftFormat.CSV:

        for i, line in enumerate(lines):
            if i == 0:
                continue
            lines[i] = line.replace(",", " ")
    if file_format == ShiftFormat.TABULAR:
        lines = _tabular_lines_to_csv_layout(lines)

    result = []
    for line in lines:
        line = line.strip()
        if not line:
            continue

        if "NUM," in line:
            # TODO we should parse it!
            continue

        fields = line.split()

        column_indices = columns[file_format]

        chain_code_index = column_indices["chain_code"]
        sequence_index = column_indices["sequence_code"]
        residue_name_index = column_indices["residue_name"]
        atom_name_index = column_indices["atom_name"]
        shift_index = column_indices["shift"]

        chain_code = fields[chain_code_index] if chain_code_index else cli_chain_code
        chain_code = chain_code if chain_code else "A"

        sequence_code = fields[sequence_index]
        sequence_code = int(sequence_code) if is_int(sequence_code) else sequence_code

        residue_name = fields[residue_name_index] if residue_name_index else UNUSED
        residue_name = (
            translate_1_to_3(residue_name)[0]
            if residue_name != UNUSED and len(residue_name) == 1
            else residue_name
        )
        atom_name = fields[atom_name_index]

        shift = float(fields[shift_index])

        residue = Residue(
            chain_code=chain_code,
            sequence_code=sequence_code,
            residue_name=residue_name,
        )
        atom = AtomLabel(residue, atom_name)
        shift = ShiftData(atom=atom, value=shift)
        result.append(shift)

    return result


def _tabular_lines_to_csv_layout(lines):

    headers = None
    result = []
    next_line_read_headers = False

    for i, line in enumerate(lines):

        line = line.strip()

        if not line:
            continue

        if line in ["BACKBONE ATOMS", "SIDECHAIN CARBON", "SIDECHAIN PROTON"]:
            next_line_read_headers = True
            continue

        if next_line_read_headers:
            headers = {header: index for index, header in enumerate(line.split())}
            atom_names = [header for header in headers if header not in ["Num", "RES"]]
            next_line_read_headers = False
            continue

        fields = line.split()
        for atom_name in atom_names:
            sequence_code = fields[headers["Num"]]
            residue_name_1_let = fields[headers["RES"]]
            residue_name_3_let = translate_1_to_3(residue_name_1_let)[0]
            shift = fields[headers[atom_name]]
            if shift == "****":
                continue

            new_line = f"{sequence_code} {residue_name_3_let} {atom_name} {shift}"
            result.append(new_line)

    return result


def _get_shifts_from_server(
    pdb_file_or_code, chain_code, cli_chain_code, use_file=False
):

    data = {
        "deuterate": 1,
        "ph": 7,
        "temper": 298,
        "phospho": 1,
        "shifttype": 0,
        "format": 3,
        "shifty": 0,
        "pdbid": "",
        "nonoverlap": 1,
    }
    files = {}
    fh = None
    if use_file:
        fh = open(pdb_file_or_code, "rb")
        files = {"file": fh}
    else:
        pdb_file_or_code = (
            f"{pdb_file_or_code}{chain_code}" if chain_code else pdb_file_or_code
        )
        data["pdbid"] = pdb_file_or_code

    if use_file:
        r = requests.request("POST", CGI_URL, data=data, files=files)

    else:
        r = requests.request("POST", CGI_URL, data=data)

    if fh:
        fh.close()

    _exit_if_bad_html_request(pdb_file_or_code, r)

    _exit_if_error_calculating_shifts(pdb_file_or_code, r)

    soup = BeautifulSoup(r.text, features="html.parser")
    shifts = ""
    putative_links = soup.find_all("a", href=True, string="download predictions")
    if putative_links:
        link = putative_links[0]["href"]
        data_url = f"{ROOT_URL}/{link}"
        data_r = requests.get(data_url)
        text = data_r.text

        shifts = _parse_text_to_shifts(text, cli_chain_code, CGI_URL)

    return shifts


def _exit_if_bad_html_request(pdb_code, r):
    if r.status_code != 200:
        msg = f"""
        failed to download shifts from {CGI_URL} for pdb code {pdb_code}
        is you network connection working?
        """
        exit_error(msg)


def _exit_if_error_calculating_shifts(pdb_code, r):

    if "Error in Calculating Chemical Shifts" in r.text:
        soup = BeautifulSoup(r.text, features="html.parser")
        texts = soup.findAll(string=True)
        visible_texts = filter(tag_visible, texts)
        text = " ".join(t.strip() for t in visible_texts)
        text = _spaces_to_single(text)
        text = _dedent_all(text)
        text = _select_text_between(text, "SHIFTX2 PREDICTIONS", "Please try again")
        msg = f"""
        failed to calculate shifts from {CGI_URL} for pdb code {pdb_code}
        the msg was:

        {text}
        """
        exit_error(msg)


def _exit_if_pdb_code_bad(pdb_code):
    if len(pdb_code) != 4:
        msg = f"""
        the code {pdb_code} is not a valid pdb code, they should be 4 letters long e.g. 1UBQ [ubiquitin]
        """
        exit_error(msg)


def _text_from_html(body):

    soup = BeautifulSoup(body, "html.parser")
    texts = soup.findAll(text=True)
    visible_texts = filter(tag_visible, texts)
    return " ".join(t.strip() for t in visible_texts)


def _select_text_between(text, start, end):
    start_index = text.find(start)
    start_index = start_index + len(start) if start != -1 else 0

    end_index = text.find(end)

    return text[start_index:end_index]


def _dedent_all(text):
    lines = []
    for line in text.split("\n"):
        lines.append(line.lstrip())

    return "\n".join(lines)


def _spaces_to_single(text):
    whitespace = re.compile(r"\s+")
    return " ".join(whitespace.split(text))


def _get_filename_from_url(url=None):
    if url is None:
        return None
    urlpath = urlsplit(url).path
    return os.path.basename(urlpath)


def _exit_if_alphafold_network_bad(alphafold_result):
    if not alphafold_result.network_ok:
        msg = f"""
           failed to get alphafold structure url using uniprot id {alphafold_result.uniprot_id} which was
           mapped from pdb code {alphafold_result.pdb_code}. The url
           {ALPHA_FOLD_URL_TEMPLATE.format(uniprot_id=alphafold_result.uniprot_id, alphafold_key=ALPHA_FOLD_KEY)}
           could not be accessed is there a network problem?
        """

        exit_error(msg)


def _exit_if_pdb_download_bad(pdb_file_data: PDBDownloadResult):

    msg = None
    if not pdb_file_data.network_ok:
        msg = f"""
            The alphafold structure {pdb_file_data.alphafold_id} derived from {pdb_file_data.pdb_code} with chain
            {pdb_file_data.chain_code} could not be download from the url:
            {pdb_file_data.pdb_url}
            is there a network problem
            """

    if not pdb_file_data.file_system_ok:
        msg = f"""
            The pdb file for the alphafold structure {pdb_file_data.alphafold_id} derived from {pdb_file_data.pdb_code}
            with chain  {pdb_file_data.chain_code} could not be saved to the file system at the path, is your disk full?
            """

    if msg:
        exit_error(msg)


def _pdb_code_to_alphafold_pdb_file(code_or_filename, source_chain, verbose):
    msg = [f"# alphafold: {code_or_filename}->"]

    mapping = _convert_pdb_code_to_uniprot_id(code_or_filename, source_chain)

    _exit_if_pdb_to_uniprot_network_failure(mapping)
    _exit_if_pdb_to_uniprot_mapping_fails(mapping)

    msg.append(f"{mapping.uniprot_id}->")

    alphafold_result = _get_alphafold_pdb_url_from_uniprot_id(mapping)

    _exit_if_alphafold_network_bad(alphafold_result)
    _exit_if_alphafold_uniprot_to_pdb_url_fails(alphafold_result)

    msg.append(alphafold_result.alphafold_id)

    pdb_file_data = _download_pdb_file(alphafold_result)

    _exit_if_pdb_download_bad(pdb_file_data)

    # TODO: why does this corrupt a NEF stream i printed to stdout rather than stderr...
    if verbose:
        print(
            "".join(msg),
            f"[{ALPHA_FOLD_URL_TEMPLATE.format(uniprot_id=mapping.uniprot_id, alphafold_key=ALPHA_FOLD_KEY)}]",
            file=sys.stderr,
        )

    return pdb_file_data


def _download_pdb_file(alphafold_result: AlphafoldResult) -> PDBDownloadResult:

    response = requests.get(alphafold_result.pdb_url)

    network_ok = response.status_code == NETWORK_200_OK

    data = response.text

    if data:
        pdb_filename = _get_filename_from_url(alphafold_result.pdb_url)
        tmpdirname = tempfile.mkdtemp()
        file_path = Path(tmpdirname) / pdb_filename
        with open(file_path, "w") as fp:
            fp.write(data)

    result = PDBDownloadResult(
        network_ok=network_ok,
        file_system_ok=file_path.is_file(),
        pdb_url=alphafold_result.pdb_url,
        pdb_code=alphafold_result.pdb_code,
        chain_code=alphafold_result.chain_code,
        uniprot_id=alphafold_result.uniprot_id,
        alphafold_id=alphafold_result.alphafold_id,
        pdb_file_path=file_path,
        pdb_uniprot_start=alphafold_result.pdb_uniprot_start,
        pdb_uniprot_end=alphafold_result.pdb_uniprot_end,
        pdb_start_residue=alphafold_result.pdb_start_residue,
    )
    return result


def _get_alphafold_pdb_url_from_uniprot_id(
    mapping: PdbUniprotMapping,
) -> AlphafoldResult:

    alphafold_url = ALPHA_FOLD_URL_TEMPLATE.format(
        uniprot_id=mapping.uniprot_id, alphafold_key=ALPHA_FOLD_KEY
    )

    response = requests.get(alphafold_url)

    network_ok = response.status_code == NETWORK_200_OK

    raw_data = response.json() if network_ok and response.text else None

    uniprot_id = mapping.uniprot_id
    entry = None
    alphafold_pdb_url = None
    alphafold_entry_id = None
    if raw_data:
        entry = raw_data[0]
        alphafold_pdb_url = entry["pdbUrl"]
        alphafold_entry_id = entry["entryId"]

    alphafold_result = AlphafoldResult(
        network_ok=network_ok,
        pdb_url=alphafold_pdb_url,
        uniprot_id=uniprot_id,
        pdb_code=mapping.pdb_code,
        chain_code=mapping.chain_code,
        raw_data=entry,
        alphafold_id=alphafold_entry_id,
        pdb_uniprot_start=mapping.pdb_uniprot_start,
        pdb_uniprot_end=mapping.pdb_uniprot_end,
        pdb_start_residue=mapping.pdb_start_residue,
    )

    return alphafold_result


def _exit_if_alphafold_uniprot_to_pdb_url_fails(alphafold_result):
    if not alphafold_result.pdb_url:
        msg = f"""
           failed to get alphafold structure url using uniprot id {alphafold_result.uniprot_id} which was
           mapped from pdb code {alphafold_result.pdb_code}. Is thisfor a structure not in the embl database
           [eg a virus, or a bacterium like E.coli]?
           """
        exit_error(msg)


def _convert_pdb_code_to_uniprot_id(
    code_or_filename, source_chain
) -> PdbUniprotMapping:

    response = requests.get(
        PDB_UNIPROT_MAPPING_URL_TEMPLATE.format(code_or_filename=code_or_filename)
    )

    network_ok = response.status_code == NETWORK_200_OK

    raw_mapping_data = response.json() if network_ok and response.text else None

    uniprot_id = None
    if raw_mapping_data:
        code_or_filename = code_or_filename.lower()

        for putative_uniprot_id, mapping_element in raw_mapping_data[code_or_filename][
            "UniProt"
        ].items():
            for mapping in mapping_element["mappings"]:
                if mapping["chain_id"] == source_chain:
                    uniprot_id = putative_uniprot_id
                    break

    pdb_uniprot_start = None
    pdb_uniprot_end = None
    pdb_start_residue = None
    if uniprot_id:

        chain_info = raw_mapping_data[code_or_filename]["UniProt"][uniprot_id]
        first_mapping = next(iter(chain_info["mappings"]))

        pdb_uniprot_start = int(first_mapping["unp_start"])
        pdb_uniprot_end = int(first_mapping["unp_end"])
        pdb_start_residue = int(
            first_mapping["start"]["residue_number"]
        )  # as opposed to 'author_residue_number'

    result = PdbUniprotMapping(
        network_ok=network_ok,
        pdb_code=code_or_filename,
        chain_code=code_or_filename,
        uniprot_id=uniprot_id,
        pdb_uniprot_start=pdb_uniprot_start,
        pdb_uniprot_end=pdb_uniprot_end,
        pdb_start_residue=pdb_start_residue,
        raw_data=raw_mapping_data,
    )

    return result


def _exit_if_pdb_to_uniprot_network_failure(mapping: PdbUniprotMapping):

    if not mapping.network_ok:
        msg = f"""
        failed to get uniprot id from pdb code {mapping.pdb_code}
        it appears there was a network problem accessing the url:

        {PDB_UNIPROT_MAPPING_URL_TEMPLATE.format(code_or_filename=mapping.pdb_code)}
        """
        exit_error(msg)


def _exit_if_pdb_to_uniprot_mapping_fails(mapping: PdbUniprotMapping):

    if mapping.uniprot_id is None:
        mapping_table = []
        for putative_uniprot_id, mapping_element in mapping.raw_data[mapping.pdb_code][
            "UniProt"
        ].items():
            for mapping in mapping_element["mappings"]:
                mapping_table.append(
                    (
                        putative_uniprot_id,
                        mapping["chain_id"],
                        mapping["unp_start"],
                        mapping["unp_end"],
                    )
                )

        headings = [
            "uniprot id",
            "chain code",
            "sequence code start",
            "sequence code end",
        ]
        mapping_table = tabulate(mapping_table, headers=headings)

        msg = """
            using pdb code {code_or_filename} and chain {source_chain} I couldn't find a mapping,
            the available mappings are

            {mapping_table}
            """
        msg = dedent(msg)
        msg = msg.format(
            code_or_filename=mapping.pdb_code.upper(),
            source_chain=mapping.chain_code,
            mapping_table=mapping_table,
        )
        exit_error(msg)



import pytest
import typer
from pynmrstar import Entry
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    read_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.pales.importers.rdcs import rdcs

runner = CliRunner()
app = typer.Typer()
app.command()(rdcs)


# noinspection PyUnusedLocal
@pytest.mark.parametrize(
    "test_name",
    [
        "anA",
        "ssiaA",
        "ssiaF",
        "anDC_short",
        "dadrFixed_short",
        "dadrOnlyFixed_short",
        "saupePred_short",
        "ssiaB_short",
        "ssiaC_short",
        "ssia_short",
        "svd_short",
        "1AKI_pales_short",
    ],
)
def test_rdcs(test_name):

    test_data_path = path_in_test_data(
        __file__,
        f"{test_name}.tbl",
    )
    expected_data = read_test_data(f"{test_name}.nef", __file__)

    result = run_and_report(
        app,
        [
            test_data_path,
        ],
    )

    expected_entry = Entry.from_string(expected_data)

    for frame in expected_entry:
        test_frame = isolate_frame(result.stdout, frame.name)
        # TODO we should be able to compare the frames and loops logically and directly without going to strings
        assert_lines_match(str(frame), test_frame)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
# import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(app, name="rpf", help="- write rpf shifts")

    # app.add_typer(import_app, name="import", help="- import fasta sequences")
    app.add_typer(export_app, name="export", help="- export rpf shifts")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.rpf.exporters.shifts  # noqa: F401

    # import nef_pipelines.transcoders.fasta.importers.sequence  # noqa: F401



import re
from pathlib import Path

import typer

from nef_pipelines.lib.test_lib import read_test_data, run_and_report
from nef_pipelines.transcoders.mars.exporters.input import input_

app = typer.Typer()
app.command()(input_)

ROOT_INPUT_FILENAME = "Sec5Part2_read_only_new_x31m_f1"


def test_input_stdout(tmpdir):
    input = read_test_data("sec5_short.neff", __file__)

    result = run_and_report(app, "-".split(), input=input)

    entry_name = "Sec5Part2_read_only_new_x31m_f1"

    INPUTS_AND_SUFFIXES = {
        "fixConn:": "_fix_con.tab",
        "fixAss:": "_fix_ass.tab",
        "sequence:": ".fasta",
        "secondary:": "_psipred.tab",
        "csTab:": "_shifts.tab",
    }

    for input, suffix in INPUTS_AND_SUFFIXES.items():
        test = re.compile(rf"{input}\s+{entry_name}{suffix}")
        assert test.search(result.stdout)

    assert re.search(r"deuterated:\s+0", result.stdout)
    assert re.search(r"rand_coil:\s+0", result.stdout)

    fix_con_path = tmpdir / f"{entry_name}_fix_con.tab"
    assert not fix_con_path.exists()

    fix_ass_path = tmpdir / f"{entry_name}_fix_ass.tab"
    assert not fix_ass_path.exists()


def test_input_stdout_deuterated_random_coil():
    input = read_test_data("sec5_short.neff", __file__)

    result = run_and_report(app, "--random-coil -".split(), input=input)

    assert re.search(r"rand_coil:\s+1", result.stdout)
    assert re.search(r"deuterated:\s+0", result.stdout)

    result = run_and_report(app, "--deuterated -".split(), input=input)

    assert re.search(r"rand_coil:\s+0", result.stdout)
    assert re.search(r"deuterated:\s+1", result.stdout)


def test_output_to_dir(tmpdir):
    input = read_test_data("sec5_short.neff", __file__)

    run_and_report(
        app,
        [
            str(tmpdir),
        ],
        input=input,
    )

    tmpdir = Path(tmpdir)
    input_path = tmpdir / f"{ROOT_INPUT_FILENAME}.inp"

    assert input_path.is_file()
    assert input_path.stat().st_size > 2023

    fix_con_path = tmpdir / f"{ROOT_INPUT_FILENAME}_fix_con.tab"

    assert fix_con_path.is_file()
    assert fix_con_path.stat().st_size == 0

    fix_ass_path = tmpdir / f"{ROOT_INPUT_FILENAME}_fix_con.tab"

    assert fix_ass_path.is_file()
    assert fix_ass_path.stat().st_size == 0


def test_output_to_dir_existing_no_force(tmpdir):
    input = read_test_data("sec5_short.neff", __file__)

    tmpdir = Path(tmpdir)
    input_path = tmpdir / f"{ROOT_INPUT_FILENAME}.inp"
    with input_path.open("w") as fh:
        fh.write("test")

    run_and_report(
        app,
        [
            str(tmpdir),
        ],
        input=input,
        expected_exit_code=1,
    )


def test_output_to_dir_existing_force(tmpdir):
    input = read_test_data("sec5_short.neff", __file__)

    tmpdir = Path(tmpdir)
    input_path = tmpdir / f"{ROOT_INPUT_FILENAME}.inp"
    write_test_file(input_path)

    run_and_report(
        app,
        [
            "--force",
            str(tmpdir),
        ],
        input=input,
    )

    assert input_path.is_file()
    assert input_path.stat().st_size > 2023


def write_test_file(input_path):
    with input_path.open("w") as fh:
        fh.write("test")


def test_fix_con_fix_ass_existing(tmpdir):
    input = read_test_data("sec5_short.neff", __file__)

    tmpdir = Path(tmpdir)

    fix_con_path = tmpdir / f"{ROOT_INPUT_FILENAME}_fix_con.tab"
    write_test_file(fix_con_path)

    fix_ass_path = tmpdir / f"{ROOT_INPUT_FILENAME}_fix_ass.tab"
    write_test_file(fix_ass_path)

    run_and_report(
        app,
        [
            str(tmpdir),
        ],
        input=input,
    )

    assert fix_con_path.stat().st_size == 4
    assert fix_ass_path.stat().st_size == 4



from nef_pipelines.lib.util import exit_error


def _exit_if_no_frames_selected(frames):
    if len(frames) == 0:
        exit_error("no shift frames selected")


def _filter_shifts_by_chain(shifts, chain):
    return [shift for shift in shifts if shift.atom.residue.chain_code == chain]


def _exit_no_chain_selected(chains_codes):
    if len(chains_codes) == 0:
        exit_error("no chain was selected, please specify a chain with -c")


def _select_target_chain_from_sequence_if_not_defined(target_chain, shift_chains):
    if target_chain is None:
        num_shift_chains = len(shift_chains)
        if num_shift_chains == 1:
            target_chain = shift_chains[0]
        elif num_shift_chains != 0:
            _exit_no_chain_selected(shift_chains)
    return target_chain



from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Tuple, Union

from strenum import StrEnum

from nef_pipelines.lib.isotope_lib import Isotope

FAKE_SPECTROMETER_FREQUENCY_600 = 600.12345678


class NMRMassClass(Enum):
    IDP = auto()
    LOW = auto()
    HIGH = auto()


class ExperimentType(StrEnum):
    N_HSQC = auto()
    C_HSQC = auto()

    HNCA = auto()
    HNcoCA = auto()

    HNCACB = auto()
    HNcoCACB = auto()
    CBCAcoNH = auto()

    HNCO = auto()
    HNcaCO = auto()

    # for transfer pathways in 13C detected spectra see
    # Kosol s. et. al Molecules 2013, 18, 10802-10828; doi:10.3390/molecules180910802 [in references]
    # Felli I.C. et. al. Chem. Rev. 2022, 122, 9468−9496 doi:10.1021/acs.chemrev.1c00871

    CON = auto()
    NCO = auto()
    CAN = auto()
    CACO = auto()
    CBCACO = auto()
    CACON = auto()
    CBCACON = auto()
    caNCO = auto()
    CANCO = auto()
    COCON = auto()
    NcaNCO = auto()
    CBCANCO = auto()
    COCA = auto()

    # NcaNCO = auto()

    CDETECT = auto()
    TRIPLE = auto()


SPECTRUM_TYPE_TO_CLASSIFICATION = {
    ExperimentType.N_HSQC: "H[N]",
    ExperimentType.C_HSQC: "H[C]",
    ExperimentType.HNCA: "H[N[CA]]",
    ExperimentType.HNcoCA: "H[N[co[CA]]]",
    ExperimentType.HNCACB: "H[N[CA[CB]]]",  # even though you might think of it as H[N[{CA|CB}]]'
    # or H[N[{CA|ca[Cali]}]]]
    # or even H[N[{CA|ca[CB]}]]
    # or even H[N[{CA|ca[C]}]]  ?
    ExperimentType.HNcoCACB: "H[N[co[CA[CB]]]]",  # even though you might think of it related to ones above
    # or 'H[N[co[{CA|ca[C]}]]]
    ExperimentType.CBCAcoNH: "h{CA|Cca}coNH",  # and why is this not h{CA|CBca}coNH
    ExperimentType.HNCO: "H[N[CO]]",
    ExperimentType.HNcaCO: "H[N[ca[CO]]]",
    ExperimentType.CON: "CON",
    ExperimentType.NCO: "NCO",
    ExperimentType.CAN: "CAN",
    ExperimentType.CACO: "CACO",
    ExperimentType.CBCACO: "CBCACO",
    ExperimentType.CACON: "CACON",
    ExperimentType.CBCACON: "CBCACON",
    ExperimentType.caNCO: "caNCO",
    ExperimentType.CANCO: "CANCO",
    ExperimentType.COCON: "CO_CO[N].Jcoupling",  # is this right
    ExperimentType.COCA: "COCA",
    ExperimentType.NcaNCO: "hNcaNCO",
    ExperimentType.CBCANCO: "CBCANCO",  # not actually defined
}

# a subset of the synonyms produced by lib/experiments/prototypes from the ccpn v2 experiment prototypes
# just for triple reoenance spectra, carbon dectected spectra and hsqcs
EXPERIMENT_SYNONYM_TO_CLASSIFICATION = {
    "15N HSQC/HMQC": "H[N]",
    "13C HSQC/HMQC": "H[C]",
    "H-detected HNcoCA": "H[N[co[CA]]]",
    "HNCA": "H[N[CA]]",
    "H-detected HNCACB": "H[N[CA[CB]]]",
    "H-detected HNcoCACB": "H[N[co[CA[CB]]]]",
    "hbCB/haCAcoNNH": "h{CA|Cca}coNH",
    "HNCO": "H[N[CO]]",
    "H-detected HNcaCO": "H[N[ca[CO]]]",
    "CON": "CON",
    "NCO": "NCO",
    "CAN": "CAN",
    "CACO": "CACO",
    "CBCACO": "CBCACO",
    "CACON": "CACON",
    "CBCACON": "CBCACON",
    "COCON": "CO_CO[N].Jcoupling",
    "COCA": "COCA",
    "caNCO": "caNCO",
    "CANCO": "CANCO",
    "NcaNCO": "hNcaNCO",
    "CBCANCO": "CBCANCO",  # not actually defined
}

EXPERIMENT_CLASSIFICATION_TO_SYNONYM = {
    synonym: spectrum_type
    for spectrum_type, synonym in EXPERIMENT_SYNONYM_TO_CLASSIFICATION.items()
}

C_DETECT_SPECTRA = {
    ExperimentType.CON,
    ExperimentType.NCO,
    ExperimentType.CAN,
    ExperimentType.CACO,
    ExperimentType.CBCACO,
    ExperimentType.CACON,
    ExperimentType.CBCACON,
    ExperimentType.COCON,
    ExperimentType.caNCO,
    ExperimentType.CANCO,
    ExperimentType.COCON,
    ExperimentType.NcaNCO,
    ExperimentType.CBCANCO,
    ExperimentType.COCA,
}

TRIPLE_SPECTRA = {
    ExperimentType.N_HSQC,
    ExperimentType.HNCA,
    ExperimentType.HNcoCA,
    ExperimentType.HNCACB,
    ExperimentType.HNcoCACB,
    ExperimentType.HNCO,
    ExperimentType.HNcaCO,
}

EXPERIMENT_GROUPS = {
    ExperimentType.CDETECT: C_DETECT_SPECTRA,
    ExperimentType.TRIPLE: TRIPLE_SPECTRA,
}


@dataclass
class PeakInfo:
    name: str
    atoms: Tuple[Union[str, Tuple[str, ...]], ...]
    dimensions: Tuple[Isotope, ...]
    atom_sets: Tuple[Tuple[Tuple[int, str], ...], ...]
    atom_set_signs: Tuple[int, ...]
    essential_atom_sets: Tuple[int, ...]
    atom_sets_by_mass: Dict[NMRMassClass, Tuple[int, ...]]
    atom_set_names: Tuple[str, ...]
    partners: Tuple[ExperimentType, ...]


@dataclass
class AtomInfo:
    n_field: int
    atom_type: str


# TODO much of this is ugly!
EXPERIMENT_INFO = {
    ExperimentType.N_HSQC: PeakInfo(
        name="N-HSQC",
        atoms=("N", "H"),
        dimensions=(Isotope.N15, Isotope.H1),
        atom_sets=(((0, "N"), (0, "H")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("HN",),
        partners=(),
    ),
    # SpectraTypes.C_HSQC: ShiftInfo(
    #     atoms={'H', 'C'},
    #     dimensions= (Isotope.H1, Isotope.N15),
    #     atom_sets=[[(1, 'H'), (1, 'C')],],
    #     atom_set_signs=[1]
    # ),
    ExperimentType.HNCA: PeakInfo(
        name="HNCA",
        atoms=(("CA",), "N", "H"),
        dimensions=(Isotope.C13, Isotope.N15, Isotope.H1),
        atom_sets=(
            ((0, "CA"), (0, "N"), (0, "H")),
            ((-1, "CA"), (0, "N"), (0, "H")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA", "CA-1"),
        partners=(ExperimentType.HNcoCA,),
    ),
    ExperimentType.HNcoCA: PeakInfo(
        name="HNcoCA",
        atoms=("H", "N", ("CA",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(((0, "H"), (0, "N"), (-1, "CA")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("CA-1",),
        partners=(ExperimentType.HNCA,),
    ),
    ExperimentType.HNCACB: PeakInfo(
        name="HNCACB",
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((0, "H"), (0, "N"), (0, "CA")),
            ((0, "H"), (0, "N"), (-1, "CA")),
            ((0, "H"), (0, "N"), (0, "CB")),
            ((0, "H"), (0, "N"), (-1, "CB")),
        ),
        atom_set_signs=(1, 1, -1, -1),
        essential_atom_sets=(0, 2),
        atom_sets_by_mass={
            NMRMassClass.LOW: (0, 1, 2, 3),
            NMRMassClass.HIGH: (0, 1, 2),
        },
        atom_set_names=("CA", "CA-1", "CB", "CB-1"),
        partners=(ExperimentType.HNcoCACB, ExperimentType.CBCAcoNH),
    ),
    ExperimentType.HNcoCACB: PeakInfo(
        name="HNcoCACB",
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((0, "H"), (0, "N"), (-1, "CA")),
            ((0, "H"), (0, "N"), (-1, "CB")),
        ),
        atom_set_signs=(1, -1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA-1", "CA-1"),
        partners=(ExperimentType.HNCACB,),
    ),
    ExperimentType.CBCAcoNH: PeakInfo(
        name="CBCAcoNH",
        atoms=("H", "N", ("CA", "CB")),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((0, "H"), (0, "N"), (-1, "CA")),
            ((0, "H"), (0, "N"), (-1, "CB")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (0, 1)},
        atom_set_names=("CA-1", "CB-1"),
        partners=(ExperimentType.HNCACB,),
    ),
    ExperimentType.HNCO: PeakInfo(
        name="HNCO",
        atoms=("H", "N", ("C",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(((0, "H"), (0, "N"), (-1, "C")),),
        atom_set_signs=(1,),
        essential_atom_sets=(0,),
        atom_sets_by_mass={NMRMassClass.LOW: (0,), NMRMassClass.HIGH: (0,)},
        atom_set_names=("C-1",),
        partners=(ExperimentType.HNcaCO,),
    ),
    ExperimentType.HNcaCO: PeakInfo(
        name="HNcaCO",
        atoms=("H", "N", ("C",)),
        dimensions=(Isotope.H1, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((0, "H"), (0, "N"), (0, "C")),
            ((0, "H"), (0, "N"), (-1, "C")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.LOW: (0, 1), NMRMassClass.HIGH: (1,)},
        atom_set_names=("C", "C-1"),
        partners=(ExperimentType.HNCO,),
    ),
    # C'-i Ni
    ExperimentType.CON: PeakInfo(
        name="CON",
        atoms=("N", "C"),
        dimensions=(Isotope.N15, Isotope.C13),
        atom_sets=(((-1, "C"), (0, "N")),),
        atom_set_signs=(1,),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("C-1",),
        partners=(),
    ),
    # Ni C'-i
    ExperimentType.NCO: PeakInfo(
        name="NCO",
        atoms=("N", "C"),
        dimensions=(Isotope.N15, Isotope.C13),
        atom_sets=(((0, "N"), (-1, "C")),),
        atom_set_signs=(1,),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("C-1",),
        partners=(),
    ),
    ExperimentType.CAN: PeakInfo(
        name="CAN",
        atoms=("N", "CA"),
        dimensions=(Isotope.N15, Isotope.C13),
        atom_sets=(((0, "N"), (0, "CA")),),
        atom_set_signs=(1,),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CA",),
        partners=(),
    ),
    ExperimentType.CACO: PeakInfo(
        name="CACO",
        atoms=("C", "CA"),
        dimensions=(Isotope.C13, Isotope.C13),
        atom_sets=(((0, "CA"), (0, "C")),),
        atom_set_signs=(1,),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CA",),
        partners=(),
    ),
    ExperimentType.CBCACO: PeakInfo(
        name="CBCACO",
        atoms=(("CA", "CB"), "C"),
        dimensions=(Isotope.C13, Isotope.C13),
        atom_sets=(
            ((0, "CA"), (0, "C")),
            ((0, "CB"), (0, "C")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CA", "CB"),
        partners=(),
    ),
    # Cαi-C'i-Ni+1
    ExperimentType.CACON: PeakInfo(
        name="CACON",
        atoms=("CA", "C", "N"),
        dimensions=(Isotope.C13, Isotope.C13, Isotope.N15),
        atom_sets=(((-1, "CA"), (-1, "C"), (0, "N")),),
        atom_set_signs=(1,),
        essential_atom_sets=(1,),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CA"),
        partners=(),
    ),
    # Cαi-1 C'i-1-Ni, Cßi-1-C'i-1-Ni
    ExperimentType.CBCACON: PeakInfo(
        name="CBCACON",
        atoms=(("CA", "CB"), "C", "N"),
        dimensions=(Isotope.C13, Isotope.C13, Isotope.N15),
        atom_sets=(
            ((-1, "CA"), (-1, "C"), (0, "N")),
            ((-1, "CB"), (-1, "C"), (0, "N")),
        ),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CA", "CB"),
        partners=(),
    ),
    # Ni-Ni-C'i-1, Ni+1-Ni-C'i-1
    ExperimentType.caNCO: PeakInfo(
        name="caNCO",
        atoms=("N", "C"),
        dimensions=(Isotope.N15, Isotope.N15, Isotope.C13),
        atom_sets=(((0, "N"), (0, "N"), (-1, "C")), ((1, "N"), (0, "N"), (-1, "C"))),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("Ndiag", "N+1"),
        partners=(),
    ),
    # Cai, C’i, Ni+1 and C’i, Ni+1 ,Cai+1
    ExperimentType.CANCO: PeakInfo(
        name="CANCO",
        atoms=("CA", "N", "C"),
        dimensions=(Isotope.C13, Isotope.C13, Isotope.N15),
        atom_sets=(((0, "CA"), (1, "N"), (0, "C")), ((1, "CA"), (1, "N"), (0, "C"))),
        atom_set_signs=(1, 1),
        essential_atom_sets=(0, 1),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("Ndiag", "N+1"),
        partners=(),
    ),
    # C'-i-C'-i-Ni, C'i-1,-C'i-Ni+1, C'i+1 -C'i -Ni+1
    ExperimentType.COCON: PeakInfo(
        name="COCON",
        atoms=("N", "C"),
        dimensions=(Isotope.C13, Isotope.C13, Isotope.N15),
        atom_sets=(
            ((-1, "C"), (-1, "C"), (0, "N")),
            ((-1, "C"), (0, "C"), (1, "N")),
            ((1, "C"), (0, "C"), (1, "N")),
        ),
        atom_set_signs=(1, 1, 1),
        essential_atom_sets=(0, 1, 2),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("C-1diag", "C-1", "C+1"),
        partners=(),
    ),
    # Ni-Ni-C'i-1, Ni+1-Ni-Ci-1, Ni-1-Ni-C'i-1
    ExperimentType.NcaNCO: PeakInfo(
        name="NcaNCO",
        atoms=("N", "C"),
        dimensions=(Isotope.N15, Isotope.N15, Isotope.C13),
        atom_sets=(
            ((0, "N"), (0, "N"), (-1, "C")),
            ((1, "N"), (0, "N"), (-1, "C")),
            ((-1, "N"), (0, "N"), (-1, "C")),
        ),
        atom_set_signs=(1, 1, 1),
        essential_atom_sets=(0, 1, 2),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("N", "N+1", "N-1"),
        partners=(),
    ),
    # Cαi-C'i-1-Ni, Cßi-C'i-1-Ni, Cßi-C'i-Ni+1, Cαi- C'i-Ni+1
    ExperimentType.CBCANCO: PeakInfo(
        name="CBCANCO",
        atoms=("N", "C", ("CA", "CB")),
        dimensions=(Isotope.C13, Isotope.C13, Isotope.N15),
        atom_sets=(
            ((0, "CA"), (0, "C"), (0, "N")),
            ((0, "CB"), (0, "C"), (0, "N")),
            ((0, "CA"), (0, "C"), (1, "N")),
            ((0, "CB"), (0, "C"), (1, "N")),
        ),
        atom_set_signs=(1, 1, 1, 1),
        essential_atom_sets=(0, 1, 2, 3),
        atom_sets_by_mass={NMRMassClass.IDP: (0)},
        atom_set_names=("CAN", "CBN", "CAN+1", "CBN+1"),
        partners=(),
    ),
}



import math
from dataclasses import dataclass, field
from itertools import combinations
from statistics import stdev
from typing import Dict, List, Tuple, Union

from fyeah import f
from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe

from nef_pipelines.lib.isotope_lib import (
    CODE_TO_ISOTOPE,
    GAMMA_RATIOS,
    MAGNETOGYRIC_RATIO_1H,
)
from nef_pipelines.lib.nef_lib import (
    NEF_RELAXATION_VERSION,
    UNUSED,
    SelectionType,
    create_nef_save_frame,
    get_frame_id,
    loop_row_namespace_iter,
    select_frames,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.shift_lib import IntensityMeasurementType
from nef_pipelines.lib.util import exit_error

SERIES_DATA_CATEGORY = "_{NAMESPACE}_series_data"


@dataclass
class RelaxationSeriesValues:
    spectra: List[str] = field(default_factory=list)
    peak_ids: List[int] = field(default_factory=list)
    variable_values: List[Union[int, float]] = field(default_factory=list)
    values: List[Union[int, float, bool]] = field(default_factory=list)


def _fit_results_as_frame(
    series_frame,
    prefix,
    entry,
    fits,
    monte_carlo_errors,
    monte_carlo_value_stats,
    monte_carlo_param_values,
    noise_level,
    version_strings,
):

    spectrum_frames = _series_frame_to_spectrum_frames(series_frame, prefix, entry)

    data_id_to_spectrum_peak_ids = _series_frame_to_data_id_to_spectrum_peak_id(
        series_frame, prefix
    )

    spectrum_peak_id_to_axis_atoms = _build_spectrum_peak_id_to_axis_atoms(
        spectrum_frames
    )

    isotope_axes = spectra_to_isotope_axes(spectrum_frames)

    _exit_if_isotope_axes_are_not_unique(isotope_axes, spectrum_frames, entry.entry_id)

    isotope_frequencies = spectra_to_isotope_frequencies(spectrum_frames)

    _exit_if_isotope_frequencies_are_not_unique(
        isotope_frequencies, spectrum_frames, entry.entry_id
    )

    isotope_frequencies = _get_unique_isotope_frequencies(isotope_frequencies)

    _exit_if_no_frequencies_found(isotope_frequencies, series_frame, entry)

    field_strength = frequencies_to_field_strength(isotope_frequencies)

    axis_of_lowest_gamma = _isotope_axes_to_lowest_gamma_axis(isotope_axes)

    frame_id = get_frame_id(series_frame)

    result_frame = create_nef_save_frame(
        f"{prefix}_relaxation_list", f"{frame_id}_fitted"
    )

    result_frame.add_tag("version", NEF_RELAXATION_VERSION)
    result_frame.add_tag("experiment_type", series_frame.get_tag("experiment_type")[0])
    result_frame.add_tag("field_strength", field_strength)
    result_frame.add_tag("value_unit", "s-1")
    result_frame.add_tag("relaxation_atom_id", axis_of_lowest_gamma)
    result_frame.add_tag("ref_value", UNUSED)
    result_frame.add_tag("source", "experimental")
    result_frame.add_tag("fitting_function", "ExponentialDecay")
    result_frame.add_tag("minimizer", "leastsq")
    result_frame.add_tag("error_method", "montecarlo")
    comment = f"""
        fitting software {version_strings}
        random_seed 42
        noise_estimate {noise_level}

        source_of_noise_estimate 'replicates'
        number_of_replicates: {len(spectrum_frames)}
    """
    # error_in_noise-estimate {}
    # fit_time {}
    result_frame.add_tag("comment", comment)

    relaxation_loop = Loop.from_scratch(f"{prefix}_relaxation")
    result_frame.add_loop(relaxation_loop)

    RELAXATION_LOOP_TAGS = "index data_id data_combination_id value value_error".split()
    RELAXATION_LOOP_ATOM_TAGS = (
        "chain_code_{axis} sequence_code_{axis} residue_name_{axis} atom_name_{axis}"
    )

    all_tags = [
        *RELAXATION_LOOP_TAGS,
    ]
    for axis in range(1, len(isotope_axes) + 1):
        all_tags.append(RELAXATION_LOOP_ATOM_TAGS.format(axis=axis).split())

    relaxation_loop.add_tag(all_tags)

    data = []
    for index, (data_id, fit) in enumerate(fits.items(), start=1):
        data_row = {"index": index, "data_id": index, "data_combination_id": UNUSED}

        peak_atoms = []
        atom_key = next(iter(data_id_to_spectrum_peak_ids[data_id]))
        for axis_atoms in spectrum_peak_id_to_axis_atoms[atom_key].values():
            peak_atoms.extend(axis_atoms)

        for i, atom in enumerate(peak_atoms, start=1):
            residue = atom.residue
            data_row.update(
                {
                    f"chain_code_{i}": residue.chain_code,
                    f"sequence_code_{i}": residue.sequence_code,
                    f"residue_name_{i}": residue.residue_name,
                    f"atom_name_{i}": atom.atom_name,
                }
            )
        data.append(data_row)
        mc_error = monte_carlo_errors[data_id]["time_constant_mc_error"]
        data_row.update(
            {
                "value": f'{fit.params["time_constant"].value:.20f}',
                "value_error": f"{mc_error:.20f}",
            }
        )

    relaxation_loop.add_data(data)

    return result_frame


def _series_frame_to_data_id_to_spectrum_peak_id(series_frame, prefix):
    result = {}
    series_loop_id = f"_{prefix}_series_data"

    series_loop = (
        series_frame[series_loop_id] if series_loop_id in series_frame else None
    )
    for row in loop_row_namespace_iter(series_loop):
        spectrum_peak_id = row.nmr_spectrum_id, row.peak_id
        result.setdefault(row.data_id, set()).add(spectrum_peak_id)
    return result


def _build_spectrum_peak_id_to_axis_atoms(spectrum_frames):
    result = {}
    for spectrum_frame in spectrum_frames:
        peaks = frame_to_peaks(spectrum_frame)

        for peak in peaks:
            axis_atoms = {}
            key = spectrum_frame.name, peak.id
            result[key] = axis_atoms

            for i, shift in enumerate(peak.shifts, start=1):
                axis_atoms.setdefault(i, set()).add(shift.atom)

    return result


def _isotope_axes_to_lowest_gamma_axis(isotope_axes):
    return min(
        isotope_axes,
        key=lambda x: min(GAMMA_RATIOS[isotope] for isotope in isotope_axes[x]),
    )


def _exit_if_no_frequencies_found(isotope_frequencies, series_frame, entry):
    if not isotope_frequencies:
        msg = f"""
            no isotope frequencies found in the series frame {series_frame.name} in entry {entry.entry_id}
        """
        exit_error(msg)


def _get_unique_isotope_frequencies(isotope_frequencies):

    return {
        isotope: list(frequency)[0]
        for isotope, frequency in isotope_frequencies.items()
    }


def _series_frame_to_spectrum_frames(series_frame, prefix, entry):

    spectrum_frames = []
    for i, row in enumerate(
        loop_row_namespace_iter(
            series_frame.get_loop(f"{prefix}_series_experiment"), convert=True
        ),
        start=1,
    ):
        spectrum_frame_name = row.nmr_spectrum_id
        if spectrum_frame_name in entry.frame_dict.keys():
            spectrum_frame = entry.get_saveframe_by_name(spectrum_frame_name)
            spectrum_frames.append(spectrum_frame)

    return spectrum_frames


# TODO:  there is some duplication here!
def _select_relaxation_series_or_exit(
    entry: Entry, frame_selectors: List[str]
) -> Tuple:
    relaxation_frames = select_frames(
        entry, "nefpls_series_list", SelectionType.CATEGORY
    )
    series_frames_and_ids = OrderedSet(
        [
            (relaxation_series.name, id(relaxation_series))
            for relaxation_series in relaxation_frames
        ]
    )

    if frame_selectors:
        named_frames = []
        for frame_selector in frame_selectors:
            named_frames.extend(
                select_frames(entry, frame_selector, SelectionType.NAME)
            )

        named_frames_and_ids = OrderedSet(
            [(named_frame.name, id(named_frame)) for named_frame in named_frames]
        )
    elif len(relaxation_frames) == 1:
        relaxation_frame = relaxation_frames[0]
        frame_key = (relaxation_frame.name, id(relaxation_frame))
        named_frames_and_ids = OrderedSet(
            [
                frame_key,
            ]
        )

    selected_frames_and_ids = series_frames_and_ids.intersection(named_frames_and_ids)

    result = [
        entry.get_saveframe_by_name(frame_name)
        for frame_name, _ in selected_frames_and_ids
    ]

    return result


def _exit_if_no_series_frames_selected(active_relaxation_frames, frame_selectors):
    if not active_relaxation_frames:
        msg = f"""
            no series frames selected by the selectors
            {" ".join(frame_selectors)}
        """
        exit_error(msg)


def _exit_if_no_frame_selectors(frame_selectors):
    if len(frame_selectors) == 0:
        msg = "you must select some frames!"
        exit_error(msg)

    return None


def _exit_if_isotope_frequencies_are_not_unique(
    isotope_frequencies, series_frame, entry_id
):
    # TODO: should identify common frames and swapped ones
    for isotope, frequencies in isotope_frequencies.items():
        if len(frequencies) > 1:
            msg = f"""
                in the series frame {series_frame.name} in entry {entry_id}
                isotope {isotope} has multiple frequencies {frequencies}
                the frequencies are {', '.join(frequencies)}
                spectrometer frequencies must be consistent in a series frame
            """

            exit_error(msg)


def _exit_if_isotope_axes_are_not_unique(isotope_axes, series_frame, entry_id):

    # TODO: should identify common frames and swapped ones
    for axis, isotopes in isotope_axes.items():
        if len(isotopes) > 1:
            msg = f"""
                in the series frame {series_frame.name} in entry {entry_id}
                the isotopes {', '.join(isotopes)} appears on multiple different axes
                this variation is not compatible with the nef relaxation lists model
                the axes must be consistent in a series frame
            """

            exit_error(msg)


def spectra_to_isotope_axes(spectrum_frames):
    isotope_axes = {}
    for spectrum_frame in spectrum_frames:
        spectrum_dimension_loop = loop_row_namespace_iter(
            spectrum_frame.get_loop("nef_spectrum_dimension")
        )
        for row in spectrum_dimension_loop:
            isotope = CODE_TO_ISOTOPE.get(row.axis_code)
            isotope_axes.setdefault(row.dimension_id, set()).add(isotope)

    return isotope_axes


def spectra_to_isotope_frequencies(spectrum_frames):

    frequencies = {}
    for spectrum_frame in spectrum_frames:
        spectrum_dimension_loop = loop_row_namespace_iter(
            spectrum_frame.get_loop("nef_spectrum_dimension")
        )
        for row in spectrum_dimension_loop:
            frequencies.setdefault(row.axis_code, set()).add(row.spectrometer_frequency)

    return frequencies


def frequencies_to_field_strength(frequencies):
    fields_by_gamma_and_isotope = {}
    for isotope_code, frequency in frequencies.items():

        if isotope_code not in CODE_TO_ISOTOPE:
            continue

        isotope = CODE_TO_ISOTOPE[isotope_code]
        gamma_ratio = GAMMA_RATIOS[isotope]

        field = frequency / MAGNETOGYRIC_RATIO_1H / gamma_ratio

        fields_by_gamma_and_isotope[gamma_ratio, isotope] = field

    highest_gamma_isotope = max(fields_by_gamma_and_isotope, key=lambda x: x[0])

    return fields_by_gamma_and_isotope[highest_gamma_isotope]


def orderOfMagnitude(number):
    return math.floor(math.log(number, 10))


# TODO: move to lib
def get_loop(frame: Saveframe, loop_name: str):
    try:
        series_experiment_loop = frame.get_loop(loop_name)
    except KeyError:
        series_experiment_loop = None
    return series_experiment_loop


def _get_spectra_by_series_variable(entry, series_experiment_loop):
    from nef_pipelines.lib.nef_lib import loop_row_namespace_iter

    spectra_by_times = {}
    for i, row in enumerate(
        loop_row_namespace_iter(series_experiment_loop, convert=True), start=1
    ):
        spectrum_frame_name = row.nmr_spectrum_id
        series_variable = row.series_variable

        spectrum_frame = entry.get_saveframe_by_name(spectrum_frame_name)

        key = i, series_variable, spectrum_frame_name
        spectra_by_times[key] = spectrum_frame

    return spectra_by_times


def _exit_if_no_series_data_loops_selected(frame, series_experiment_loop):
    if not series_experiment_loop:
        raise Exception()
        msg = f"no nefpls series data loop found in frame {frame.name}"
        exit_error(msg)


def _exit_if_spectra_are_missing(spectra_by_times_and_indices, series_name):
    for (
        _,
        _,
        spectrum_frame_name,
    ), spectrum_frame in spectra_by_times_and_indices.items():
        if not spectrum_frame:
            msg = f"no spectrum frame found for series {series_name} for spectrum {spectrum_frame_name}"
            exit_error(msg)


def _get_atoms_and_values(
    peaks_by_times_and_indices, data_type: IntensityMeasurementType
) -> Dict:
    atoms_to_values = {}
    for (index, x_value, spectrum_name), peaks in peaks_by_times_and_indices.items():

        for peak in peaks:
            peak_atoms = tuple([shift.atom for shift in peak.shifts])
            x_and_y = atoms_to_values.setdefault(
                tuple(peak_atoms), RelaxationSeriesValues()
            )
            x_and_y.peak_ids.append(peak.id)
            x_and_y.variable_values.append(x_value)
            y_value = (
                peak.height
                if data_type == IntensityMeasurementType.HEIGHT
                else peak.volume
            )
            x_and_y.values.append(y_value)
            x_and_y.spectra.append(spectrum_name)

    return atoms_to_values


def _get_noise_from_duplicated_values(xy_data) -> float:

    repetitons = []

    for series in xy_data.values():
        xs = series[0]
        ys = series[1]
        ys_by_x = {}
        for x, y in zip(xs, ys):
            ys_by_x.setdefault(float(x), []).append(float(y))
        repeated_values = [values for values in ys_by_x.values() if len(values) > 1]
        repetitons.extend(repeated_values)

    differences = []
    for repetiton_set in repetitons:
        for combination in combinations(repetiton_set, 2):
            differences.append(combination[0] - combination[1])

    replicates_stdev = stdev(differences)
    replicates_stderr = replicates_stdev / len(differences) ** 0.5

    return replicates_stdev, replicates_stderr / replicates_stdev, len(differences)


def _series_frame_to_id_series_data(series_frame: Saveframe, prefix: str, entry: Entry):

    NAMESPACE = prefix  # noqa: F841
    series_category = f(SERIES_DATA_CATEGORY)
    series_data_loop = (
        series_frame.get_loop(series_category)
        if series_category in series_frame
        else None
    )

    _exit_if_no_series_data_loop(series_data_loop, series_frame, entry)

    id_to_xy_data = {}
    for row in loop_row_namespace_iter(series_data_loop):
        data_id = row.data_id
        relaxation_series = id_to_xy_data.setdefault(data_id, RelaxationSeriesValues())
        relaxation_series.spectra.append(row.nmr_spectrum_id)
        relaxation_series.peak_ids.append(row.peak_id)
        relaxation_series.variable_values.append(row.variable_value)
        relaxation_series.values.append(row.value)

    return id_to_xy_data


def _exit_if_no_series_data_loop(series_data_loop, series_frame, entry):
    if not series_data_loop:
        msg = f"""
        in the entry {entry.entry_id}
        the series frame {series_frame.name} does not have a {SERIES_DATA_CATEGORY} loop
        """
        exit_error(msg)



# TODO move some of this to a spark lib
import re
import sys
from enum import auto
from os.path import commonprefix
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from pynmrstar import Entry, Loop
from strenum import LowercaseStrEnum
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.peak_lib import frame_to_peaks
from nef_pipelines.lib.sequence_lib import TRANSLATIONS_3_1_PROTEIN
from nef_pipelines.lib.structures import NewPeak
from nef_pipelines.lib.util import (
    STDIN,
    STDOUT,
    exit_error,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.sparky import export_app

STARS_30 = "*" * 30

STDOUT_STR = str(STDOUT)


FILE_NAME_TEMPLATE_HELP = """\
    the template for the filename to export to %s will get replaced by the axis_name of the peak
    frame. - will write to stdout.
"""


class SparkyPeaksExportException(Exception):
    pass


class OverlappingChainsException(SparkyPeaksExportException):
    pass


SPECTRUM_CATEGORY = "nef_nmr_spectrum"
FULL_ASSIGNMENTS_HELP = """
    print full assignments of the form PR_36N-PR_36H [generally assignments are printed in the form PR_36N-H if all the
    the residues in the assignment are the same]
"""
ADD_DATA_HELP = "don't include the data string before the volume and height columns"

NO_CHAINS_HELP = "don't include chains in the output [this will fail warnings if try to output more than one chain...]"

NO_NEGATIVES_HELP = (
    "don't include information on shifts from negative nmr_residues "
    "[default: @-.@65-1..CA -> PR_65p1CA with this option @-.@65-1..CA -> ?CA]"
)

DISCARD_ASSIGNMENTS_HELP = (
    "discard the peak assignments. note: peaks unassign provides more options"
)


class SUPRESSABLE_COLUMNS(LowercaseStrEnum):
    ASSIGNMENT = auto()
    HEIGHT = auto()
    VOLUME = auto()


COLUMNS_TO_SUPPRESS_HELP = f"""
    name columns to suppress in the output [{', '.join(SUPRESSABLE_COLUMNS)}], can be called multiple times or with a
    comma sepatated list [no spaces!]
"""

DEFAULT_CHAIN_SEPARATOR = "."


# noinspection PyUnusedLocal
@export_app.command()
def peaks(
    file_name_template: str = typer.Option(
        "%s.txt",
        help=FILE_NAME_TEMPLATE_HELP,
        metavar="<peak-file.xpk>",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    frame_selectors: List[str] = typer.Argument(
        None, help="the names of the frames to export", metavar="<frames>"
    ),
    discard_assignments: bool = typer.Option(False, help=DISCARD_ASSIGNMENTS_HELP),
    full_assignments: bool = typer.Option(False, help=FULL_ASSIGNMENTS_HELP),
    add_data: bool = typer.Option(False, help=ADD_DATA_HELP),
    no_chains: bool = typer.Option(False, help=NO_CHAINS_HELP),
    chain_separator: str = typer.Option(
        DEFAULT_CHAIN_SEPARATOR, help="characters used to separate chains from residues"
    ),
    no_negative_residues: bool = typer.Option(False, help=NO_NEGATIVES_HELP),
    columns_to_suppress: List[str] = typer.Option(
        [], "--suppress-column", help=COLUMNS_TO_SUPPRESS_HELP
    ),
    # no_volume: bool = typer.Option(
    #     False, help="don't include a volume column in the output"
    # ),
    # no_height: bool = typer.Option(
    #     False, help="don't include a height column in the output"
    # ),
    # no_assignment: bool = Typer
):
    """-  write sparky peaks"""

    columns_to_suppress = parse_comma_separated_options(columns_to_suppress)

    output_to_files = file_name_template != STDOUT_STR

    frame_selectors = parse_comma_separated_options(frame_selectors)

    if len(frame_selectors) == 0:
        frame_selectors = [
            "*",
        ]

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    selected_frame_names = [
        frame.name for frame in select_frames_by_name(entry, frame_selectors)
    ]

    _if_no_frames_selected_exit(frame_selectors, selected_frame_names)

    _if_multiple_frame_and_no_template_exit_error(
        file_name_template, selected_frame_names
    )

    frame_name_to_file_name = _build_file_names(
        file_name_template, selected_frame_names
    )

    try:
        entry, sparky_tables = pipe(
            entry,
            selected_frame_names,
            include_assignments=not discard_assignments,
            include_full_assignments=full_assignments,
            show_data_tag_in_header=add_data,
            chain_separator=chain_separator,
            no_negative_pseudo_residues=no_negative_residues,
            no_chains=no_chains,
            columns_to_suppress=columns_to_suppress,
        )
    except SparkyPeaksExportException as e:
        exit_error(str(e))

    sparky_tables = _map_frame_names_to_paths_if_required(
        sparky_tables, frame_name_to_file_name, output_to_files
    )

    _write_output_tables(sparky_tables, output_to_files)

    _output_entry_if_required(entry, output_to_files)


def _output_entry_if_required(entry, output_to_files):

    if (not sys.stdout.isatty()) and output_to_files:
        print(entry)


def _write_output_tables(sparky_lines, output_to_files):

    if output_to_files:
        _write_results_to_files(sparky_lines)
    else:
        _print_results_to_stderr(sparky_lines)


def _map_frame_names_to_paths_if_required(
    sparky_lines, frame_name_to_file_name, output_to_files
):
    if output_to_files:
        sparky_lines = {
            frame_name_to_file_name[frame_name]: header_and_lines
            for frame_name, header_and_lines in sparky_lines.items()
        }
    return sparky_lines


def _write_results_to_files(sparky_lines):
    for file_name, header_and_lines in sparky_lines.items():
        with open(file_name, "w") as fp:
            _print_table(*sparky_lines[file_name], fp)


def _print_results_to_stderr(sparky_lines):

    for frame_name, header_and_lines in sparky_lines.items():
        header = f"{STARS_30} {frame_name} {STARS_30}"

        print(header, file=sys.stderr)

        print(file=sys.stderr)

        _print_table(*header_and_lines, sys.stderr)

        print(file=sys.stderr)

    if len(sparky_lines) > 0:
        trailer = f'{STARS_30}*{"*" * len(frame_name)}*{STARS_30}'
        print(trailer, file=sys.stderr)


def _build_file_names(file_name_template, selected_frame_names):
    if "%s" not in file_name_template and file_name_template != "-":
        return {selected_frame_names[0]: file_name_template}
    elif file_name_template == "-":
        return {frame_name: frame_name for frame_name in selected_frame_names}
    else:
        return {
            frame_name: file_name_template % frame_name
            for frame_name in selected_frame_names
        }


def _if_no_frames_selected_exit(frame_selectors, selected_frame_names):
    if len(selected_frame_names) == 0:
        msg = f"""\
            WARNING []: no spectrum frames were selected by the selectors {' '.join(frame_selectors)}
        """
        print(msg, file=sys.stderr)
        exit(0)


def _if_multiple_frame_and_no_template_exit_error(
    file_name_template, selected_frame_names
):
    if file_name_template != "-":
        if "%s" not in file_name_template and len(selected_frame_names) > 1:
            exit_error(
                f"%s is not in the filename template and there is more than one file, template {file_name_template}"
            )


def _print_table(header, lines, file_pointer):

    table = [header, [""], *lines]
    alignments = ["right"] * len(header)
    alignments[0] = "left"
    print(tabulate(table, colalign=alignments, tablefmt="plain"), file=file_pointer)


def _if_chains_overlap_raise(peaks):
    peak_residues_by_chain = {}
    for peak in peaks:
        for shift in peak.shifts:
            chain_code = shift.atom.residue.chain_code
            sequence_code = shift.atom.residue.sequence_code

            peak_residues_by_chain.setdefault(chain_code, set()).add(sequence_code)

    residue_intersection = set.intersection(*peak_residues_by_chain.values())
    if len(peak_residues_by_chain) > 1:
        msg = f"""
            you asked to ignore chains but there is an overlap in the sequence codes between the chains present
            the chain codes  are {' '.join(peak_residues_by_chain.keys())} the overlapping sequence codes are
            {' '.join([str(residue) for residue in residue_intersection])}
        """

        raise OverlappingChainsException(msg)


def pipe(
    entry: Entry,
    selected_frame_names: List[str],
    include_assignments=True,
    include_full_assignments: bool = False,
    show_data_tag_in_header=False,
    chain_separator=":",
    no_negative_pseudo_residues=False,
    no_chains=False,
    columns_to_suppress: List[SUPRESSABLE_COLUMNS] = (),
) -> Tuple[Entry, Dict[str, List[str]]]:

    selected_frame_names = set(selected_frame_names)

    spectrum_frames = entry.get_saveframes_by_category(SPECTRUM_CATEGORY)

    names_and_frames = {
        frame.name: frame
        for frame in spectrum_frames
        if frame.name in selected_frame_names
    }

    sparky_lines = {}
    for frame_name, frame in names_and_frames.items():

        peaks = frame_to_peaks(frame)

        if no_chains:
            _if_chains_overlap_raise(peaks)

        sparky_lines[frame_name] = _build_sparky_lines(
            peaks,
            include_assignments=include_assignments,
            abbreviate_assignments=not include_full_assignments,
            show_data_tag_in_header=show_data_tag_in_header,
            chain_sequence_separator=chain_separator,
            no_negative_pseudo_residues=no_negative_pseudo_residues,
            no_chains=no_chains,
            columns_to_suppress=columns_to_suppress,
        )

    return entry, sparky_lines


def _check_column_has_floats(peak_list: List[NewPeak], column_name: str):
    result = False

    for peak in peak_list:

        value = getattr(peak, column_name)
        if value != UNUSED and is_float(value):
            result = True
            break
    return result


def _peak_loop_has_volumes(loop: Loop):
    return _check_column_has_floats(loop, "volume")


def _peak_loop_has_heights(loop: Loop):
    return _check_column_has_floats(loop, "height")


def _build_sparky_lines(
    peaks_list: List[NewPeak],
    include_assignments=False,
    abbreviate_assignments=False,
    show_data_tag_in_header=False,
    chain_sequence_separator=".",
    no_negative_pseudo_residues=False,
    no_chains=False,
    columns_to_suppress=(),
):

    lines = []

    header = []
    if SUPRESSABLE_COLUMNS.ASSIGNMENT not in columns_to_suppress:
        header.append("Assignment")

    if len(peaks_list) > 0:
        num_dimensions = len(peaks_list[0].shifts)

        position_headers = [
            f"w{dimension}" for dimension in range(1, num_dimensions + 1)
        ]

        header.extend(position_headers)

        has_volumes = (
            _peak_loop_has_volumes(peaks_list)
            and SUPRESSABLE_COLUMNS.VOLUME not in columns_to_suppress
        )
        has_heights = (
            _peak_loop_has_heights(peaks_list)
            and SUPRESSABLE_COLUMNS.HEIGHT not in columns_to_suppress
        )

        if (has_heights or has_volumes) and show_data_tag_in_header:
            header.append("Data")

        if has_heights:
            header.append("Height")

        if has_volumes:
            header.append("Volume")

        #
        # # could check for line widths here
        # # format for line widths is lw1 (hz)
        #

        for peak in peaks_list:
            out_row = []
            lines.append(out_row)

            assignments = []

            number_residues = len(
                set([shift.atom.residue.sequence_code for shift in peak.shifts])
            )
            number_chains = len(
                set([shift.atom.residue.chain_code for shift in peak.shifts])
            )

            can_abbreviate = number_residues == 1 and number_chains == 1

            if SUPRESSABLE_COLUMNS.ASSIGNMENT not in columns_to_suppress:
                if include_assignments:
                    for atom in [shift.atom for shift in peak.shifts]:

                        chain_code = atom.residue.chain_code
                        sequence_code = str(atom.residue.sequence_code)
                        residue_name = atom.residue.residue_name
                        atom_name = atom.atom_name
                        current_chain_sequence_separator = chain_sequence_separator

                        if chain_code == "@-":
                            chain_code = ""
                        elif chain_code.startswith("#"):
                            chain_code = f"PC_{chain_code}"
                        elif chain_code == UNUSED or no_chains:
                            chain_code = ""
                            current_chain_sequence_separator = ""

                        if not chain_code or len(chain_code) == 0:
                            current_chain_sequence_separator = ""

                        if sequence_code.endswith("-1"):
                            if no_negative_pseudo_residues:
                                sequence_code = "?"
                            else:
                                sequence_code = f"{sequence_code[:-2]}"
                                atom_name = f"{atom_name}m1"
                            can_abbreviate = False

                        if sequence_code.startswith("@"):
                            sequence_code = f"PR_{sequence_code[1:]}"
                        # if chain_code.startswith('PC_'):
                        #     sequence_code = f'{chain_code}:{sequence_code}'
                        #     chain_sequence_separator = ''

                        if residue_name in TRANSLATIONS_3_1_PROTEIN:
                            residue_name = TRANSLATIONS_3_1_PROTEIN[residue_name]

                        if not residue_name:
                            residue_name = ""

                        group = f"{chain_code}{current_chain_sequence_separator}{residue_name}{sequence_code}"

                        assignment = f"{group}{atom_name}"
                        assignments.append(assignment)

                else:
                    assignments.extend(["?"] * num_dimensions)

                common_prefix = commonprefix(assignments)
                need_to_join = True
                if common_prefix != "?" and abbreviate_assignments and can_abbreviate:

                    split_common_prefix = re.split(r"(\d+)", common_prefix)
                    common_prefix = f"{split_common_prefix[0]}{split_common_prefix[1]}"

                    if len(common_prefix) > 0 and common_prefix != "?":
                        for i, assignment in enumerate(assignments):
                            assignments[i] = assignment[len(common_prefix) :]

                        assignments = f"{common_prefix}{'-'.join(assignments)}"
                        need_to_join = False

                if need_to_join:
                    assignments = "-".join(assignments)

                out_row.append(assignments)

            out_row.extend([f"{shift.value:7.3f}" for shift in peak.shifts])

            if (has_heights or has_volumes) and show_data_tag_in_header:
                out_row.append("")
            if has_heights:
                height = peak.height if peak.height != UNUSED else 0.000
                out_row.append(height)

            if has_volumes:
                volume = peak.volume if peak.volume != UNUSED else 0.000
                out_row.append(volume)

    return header, lines
